# pylint: disable=line-too-long
r"""

.. figure:: img/elliptical_cylinder_geometry.png

   Elliptical cylinder geometry $a = r_\text{minor}$
   and $\nu = r_\text{major} / r_\text{minor}$ is the *axis_ratio*.

The function calculated is

.. math::

    I(\vec q)=\frac{1}{V_\text{cyl}}\int{d\psi}\int{d\phi}\int{
        p(\theta,\phi,\psi)F^2(\vec q,\alpha,\psi)\sin(\alpha)d\alpha}

with the functions

.. math::

    F(q,\alpha,\psi) = 2\frac{J_1(a)\sin(b)}{ab}

where

.. math::

    a = qr'\sin(\alpha)

    b = q\frac{L}{2}\cos(\alpha)

    r'=\frac{r_{minor}}{\sqrt{2}}\sqrt{(1+\nu^{2}) + (1-\nu^{2})cos(\psi)}


and the angle $\psi$ is defined as the orientation of the major axis of the
ellipse with respect to the vector $\vec q$. The angle $\alpha$ is the angle
between the axis of the cylinder and $\vec q$.


For 1D scattering, with no preferred orientation, the form factor is averaged
over all possible orientations and normalized by the particle volume

.. math::

    P(q) = \text{scale}  <F^2> / V

For 2d data the orientation of the particle is required, described using a
different set of angles as in the diagrams below, for further details of the
calculation and angular dispersions see :ref:`orientation`.


.. figure:: img/elliptical_cylinder_angle_definition.png

    Note that the angles here are not the same as in the equations for the
    scattering function. Rotation $\theta$, initially in the $xz$ plane, is
    carried out first, then rotation $\phi$ about the $z$ axis, finally
    rotation $\Psi$ is now around the axis of the cylinder. The neutron or
    X-ray beam is along the $z$ axis.

.. figure:: img/elliptical_cylinder_angle_projection.png

    Examples of the angles for oriented elliptical cylinders against the
    detector plane, with $\Psi$ = 0.

The $\theta$ and $\phi$ parameters to orient the cylinder only appear in the
model when fitting 2d data.


NB: The 2nd virial coefficient of the cylinder is calculated based on the
averaged radius $(=\sqrt{r_\text{minor}^2 * \text{axis ratio}})$ and length
values, and used as the effective radius for $S(Q)$ when $P(Q)*S(Q)$ is applied.


Validation
----------

Validation of our code was done by comparing the output of the 1D calculation
to the angular average of the output of the 2D calculation over all possible
angles.

In the 2D average, more binning in the angle $\phi$ is necessary to get the
proper result. The following figure shows the results of the averaging by
varying the number of angular bins.

.. figure:: img/elliptical_cylinder_averaging.png

    The intensities averaged from 2D over different numbers of bins and angles.

References
----------

#. L A Feigin and D I Svergun, *Structure Analysis by Small-Angle
   X-Ray and Neutron Scattering*, Plenum, New York, (1987) [see table 3.4]
#. L. Onsager, *Ann. New York Acad. Sci.*, 51 (1949) 627-659

Authorship and Verification
----------------------------

* **Author:**
* **Last Modified by:**
* **Last Reviewed by:**  Richard Heenan - corrected equation in docs **Date:** December 21, 2016
"""

import numpy as np
from numpy import pi, inf, sqrt, sin, cos

name = "elliptical_cylinder"
title = "Form factor for an elliptical cylinder."
description = """
    Form factor for an elliptical cylinder.
    See L A Feigin and D I Svergun, Structure Analysis by Small-Angle X-Ray and Neutron Scattering, Plenum, New York, (1987).
"""
category = "shape:cylinder"

# pylint: disable=bad-whitespace, line-too-long
#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["radius_minor",     "Ang",        20.0,  [0, inf],    "volume",      "Ellipse minor radius"],
              ["axis_ratio",   "",          1.5,   [1, inf],    "volume",      "Ratio of major radius over minor radius"],
              ["length",      "Ang",        400.0, [1, inf],    "volume",      "Length of the cylinder"],
              ["sld",         "1e-6/Ang^2", 4.0,   [-inf, inf], "sld",         "Cylinder scattering length density"],
              ["sld_solvent", "1e-6/Ang^2", 1.0,   [-inf, inf], "sld",         "Solvent scattering length density"],
              ["theta",       "degrees",    90.0,  [-360, 360], "orientation", "cylinder axis to beam angle"],
              ["phi",         "degrees",    0,     [-360, 360], "orientation", "rotation about beam"],
              ["psi",         "degrees",    0,     [-360, 360], "orientation", "rotation about cylinder axis"]]

# pylint: enable=bad-whitespace, line-too-long

source = ["lib/polevl.c", "lib/sas_J1.c", "lib/gauss76.c", "elliptical_cylinder.c"]
have_Fq = True
radius_effective_modes = [
    "equivalent cylinder excluded volume",
    "equivalent volume sphere", "average radius", "min radius", "max radius",
    "equivalent circular cross-section",
    "half length", "half min dimension", "half max dimension", "half diagonal",
    ]

def random():
    """Return a random parameter set for the model."""
    # V = pi * radius_major * radius_minor * length;
    volume = 10**np.random.uniform(3, 9)
    length = 10**np.random.uniform(1, 3)
    axis_ratio = 10**np.random.uniform(0, 2)
    radius_minor = sqrt(volume/length/axis_ratio)
    volfrac = 10**np.random.uniform(-4, -2)
    pars = dict(
        #background=0, sld=0, sld_solvent=1,
        scale=1e9*volfrac/volume,
        length=length,
        radius_minor=radius_minor,
        axis_ratio=axis_ratio,
    )
    return pars

q = 0.1
# april 6 2017, rkh added a 2d unit test, NOT READY YET pull #890 branch assume correct!
qx = q*cos(pi/6.0)
qy = q*sin(pi/6.0)

tests = [
    #[{'radius_minor': 20.0, 'axis_ratio': 1.5, 'length':400.0}, 'ER', 79.89245454155024],
    #[{'radius_minor': 20.0, 'axis_ratio': 1.2, 'length':300.0}, 'VR', 1],

    # The SasView test result was 0.00169, with a background of 0.001
    [{'radius_minor': 20.0, 'axis_ratio': 1.5, 'sld': 4.0, 'length':400.0,
      'sld_solvent':1.0, 'background':0.0},
     0.001, 675.504402],
    #[{'theta':80., 'phi':10.}, (qx, qy), 7.88866563001 ],
]



r"""
Definition
----------

This model calculates the SAS signal of a phase separating system undergoing
spinodal decomposition. The scattering intensity $I(q)$ is calculated as

.. math::
    I(q) = I_{max}\frac{(1+\gamma/2)x^2}{\gamma/2+x^{2+\gamma}}+B

where $x=q/q_0$, $q_0$ is the peak position, $I_{max}$ is the intensity
at $q_0$ (parameterised as the $scale$ parameter), and $B$ is a flat
background. The spinodal wavelength, $\Lambda$, is given by $2\pi/q_0$.

The definition of $I_{max}$ in the literature varies. Hashimoto *et al* (1991)
define it as

.. math::
    I_{max} = \Lambda^3\Delta\rho^2

whereas Meier & Strobl (1987) give

.. math::
    I_{max} = V_z\Delta\rho^2

where $V_z$ is the volume per monomer unit.

The exponent $\gamma$ is equal to $d+1$ for off-critical concentration
mixtures (smooth interfaces) and $2d$ for critical concentration mixtures
(entangled interfaces), where $d$ is the dimensionality (ie, 1, 2, 3) of the
system. Thus 2 <= $\gamma$ <= 6. A transition from $\gamma=d+1$ to $\gamma=2d$
is expected near the percolation threshold.

As this function tends to zero as $q$ tends to zero, in practice it may be
necessary to combine it with another function describing the low-angle
scattering, or to simply omit the low-angle scattering from the fit.

References
----------

#. H. Furukawa. Dynamics-scaling theory for phase-separating unmixing mixtures:
   Growth rates of droplets and scaling properties of autocorrelation functions.
   *Physica A* 123, 497 (1984).
#. H. Meier & G. Strobl. Small-Angle X-ray Scattering Study of Spinodal
   Decomposition in Polystyrene/Poly(styrene-co-bromostyrene) Blends.
   *Macromolecules* 20, 649-654 (1987).
#. T. Hashimoto, M. Takenaka & H. Jinnai. Scattering Studies of Self-Assembling
   Processes of Polymer Blends in Spinodal Decomposition.
   *J. Appl. Cryst.* 24, 457-466 (1991).

Authorship and Verification
----------------------------

* **Author:** Dirk Honecker **Date:** Oct 7, 2016
* **Last Modified by:** Steve King **Date:** Oct 25, 2018
* **Last Reviewed by:** Steve King **Date:** Oct 25, 2018
"""

import numpy as np
from numpy import inf, errstate

name = "spinodal"
title = "Spinodal decomposition model"
description = """\
      I(q) = Imax ((1+gamma/2)x^2)/(gamma/2+x^(2+gamma)) + background

      List of default parameters:

      Imax = correlation peak intensity at q_0
      background = incoherent background
      gamma = exponent (see model documentation)
      q_0 = correlation peak position [1/A]
      x = q/q_0"""

category = "shape-independent"

# pylint: disable=bad-whitespace, line-too-long
#             ["name", "units", default, [lower, upper], "type", "description"],
parameters = [["gamma",      "",    3.0, [-inf, inf], "", "Exponent"],
              ["q_0",  "1/Ang",     0.1, [-inf, inf], "", "Correlation peak position"]
             ]
# pylint: enable=bad-whitespace, line-too-long

def Iq(q,
       gamma=3.0,
       q_0=0.1):
    """
    :param q:              Input q-value
    :param gamma:          Exponent
    :param q_0:            Correlation peak position
    :return:               Calculated intensity
    """
    with errstate(divide='ignore'):
        x = q/q_0
        inten = ((1 + gamma / 2) * x ** 2) / (gamma / 2 + x ** (2 + gamma))
    return inten
Iq.vectorized = True  # Iq accepts an array of q values

def random():
    """Return a random parameter set for the model."""
    pars = dict(
        scale=10**np.random.uniform(1, 3),
        gamma=np.random.uniform(0, 6),
        q_0=10**np.random.uniform(-3, -1),
    )
    return pars



#!/usr/bin/env python
"""
Submit a batch fit job to the slurm cluster.

Given a model.py file defining a Bumps problem defined on a single data
file, with the data file specified as a command line argument, run the
bumps fit as a batch over a set of different datafiles independently.
An example model is given in model_ellipsoid_hayter_msa.py, which fits 
the data in 09319*.dat.

To run the fit, use::

    slurm_batch.py [--slurm_opts] model.py *.dat --store=T1 [--bumps_opt ...]

For example::

    slurm_batch.py model_ellipsoid_hayter_msa.py 09319*.dat --store=T1

You may need to run in a particular python environment on the 
compute nodes::

    slurm_batch.py --python=path/to/env/bin/python ...

This creates the T1 subdirectory to hold the fit results and 
prints the real command that is submitted, as well as the job id.

The store directory T1 contains a copy of the model file and 
all the data files.  The fit results for each file will be 
in T1/##/*.  The file T1/files.dat contains the list 
of "subdirectory filename" pairs indicating which ## directory 
contains the resuls for which file.  Check for errors using::

    cat T1/slurm*_1.out

The following slurm options are used::

    --array=1-#files     batch size comes from the file list
    --gres=gpu:1         request a gpu for each fit
    --job-name=model.py  use model file name for job name
    --output=...         log into T1/slurm-job_##.out
    --chdir=...          run fit from store directory
    --time=2             time as number of hours (can override)

To receive an email on job completion or failure, add the following
slurm options before the model file::

    --mail-type=END,FAIL --mail-user=user@mail.domain

Bumps options are described at bumps.readthedocs.org, with the
following set automatically::

    --batch              run in batch mode, without output to .mon
    --view=log           SAS fits want log plots
    --time=2-0.1         slurm time minus 6 minutes for cleanup

The --store and --resume options indicate the parent directory for
the output.  These are modified to store the results in a separate
subdirectory for each file.  Keep in mind that the fit is run from
the store directory, so any files or modules referenced from the
model file will need to use a full path to the original location.

After submitting the job a job id will be printed to the console.
You can check the status of the job using the usual slurm commands
such as::

    squeue

or cancel the job using::

    scancel jobid

The slurm_batch program runs directly from the source tree for sasmodels, 
and requires sasview, bumps and periodictable as sister directories
accessible on the worker nodes.  You can link it into your bin directory
using::

     mkdir ~/bin
     ln -s path/to/slurm_batch.py ~/bin

or if you are a cluster administrator, into /usr/local/bin.
"""

# If called from command line, this submits a job to the slurm queue, with _this_ file
# as the batch script. Before calling it on the worker node, slurm sets the
# SLURM_ARRAY_TASK_ID to the current task so we can tell that we are running
# as a worker and which file we should be working on.

## SBATCH options as comments do not seem to work.  Maybe they neeed to be before 
## the doc string?  For now they are hardcoded in the sbatch call in submit_job.

import sys
import os
import tempfile
import shutil

DEFAULT_TIME_LIMIT = 2

def split_args():
    slurm_opts = []
    bumps_opts = []
    model_file = None
    store = None
    resume = None
    data_files = []
    time_limit = DEFAULT_TIME_LIMIT
    interpreter = sys.executable

    # start with '-' arguments as slurm opts, then after
    # the model file any '-' arguments are bumps opts.
    opts = slurm_opts
    for v in sys.argv[1:]:
        if v.startswith('--store='):
            store = os.path.realpath(os.path.abspath(v[8:]))
        elif v.startswith('--resume='):
            resume = os.path.realpath(os.path.abspath(v[9:]))
        elif v.startswith('--time='):
            time_limit = float(v[7:])
        elif v.startswith('--python='):
            interpreter = v[9:]
        elif v[0] == '-':
            opts.append(v)
        elif model_file is None:
            model_file = v
            opts = bumps_opts
        else:
            data_files.append(v)


    s = time_limit*3600
    slurm_opts.append("--time=%d:%02d:%02d"%(s//3600, (s%3600)//60, s%60))
    bumps_opts.append('--time=%f'%(time_limit - 0.1))  # 6 min to stop cleanly

    return {
        'python': interpreter,
        'slurm': slurm_opts, 
        'model_file': model_file, 
        'data_files': data_files, 
        'store': store, 
        'resume': resume,
        'bumps': bumps_opts,
    }

def dirn(path, n):
    path = os.path.realpath(os.path.abspath(path))
    for _ in range(n):
        path = os.path.dirname(path)
    return path

def submit_job():
    # sbatch --array=1-5 ./slurm_batch.py model_ellipsoid_hayter_msa.py 09*.dat --store=T1 --fit=dream
    opts = split_args()
    store = opts['store']
    model_file = opts['model_file']
    data_files = opts['data_files']
    bumps_opts = opts['bumps']
    slurm_opts = opts['slurm']
    interpreter = opts['python']

    # make sure the store directory exists and save the order of the files, as well
    # as the model and the data files
    if store is not None:
        if not os.path.exists(store):
            os.makedirs(store)

        # save file order
        with open(os.path.join(store, 'files.dat'), 'w') as fid:
            for k, f in enumerate(data_files):
                fid.write("%02d %s\n"%(k+1, f))

        # Copy the model and data files to the root store directory
        # Since bumps changes into the model directory prior to loading
        # the datafiles, strip all leading paths from data and model and
        # set the working directory for the job to the store directory.
        model_copy = os.path.basename(model_file)
        shutil.copy(model_file, os.path.join(store, model_copy))
        data_copy = []
        for f in data_files:
            f_copy = os.path.basename(f)
            shutil.copy(f, os.path.join(store, f_copy))
            data_copy.append(f_copy) 

        model_file = model_copy
        data_files = data_copy


    # build and run the command
    SRC = dirn(__file__, 3) # __file__ is $SRC/sasmodels/example/slurm_batch.py
    parts = [
        "sbatch",
        "--array=1-%d"%len(data_files),
        "--gres=gpu:1",
        "--job-name="+model_file,
        ## since we are setting the current working directory, we don't need
        ## to fiddle the slurm output files
        "--output=%s/slurm-%%A_%%a.out"%store,
        "--chdir=%s"%store,
        ]
    parts.extend(slurm_opts)
    parts.append(__file__)
    # Remember the source root so we can reconstruct the correct python path
    # This is done after the model file so that it doesn't get interpreted
    # as a slurm option.
    parts.append("--python=%s"%interpreter)
    parts.append("--source_root=%s"%SRC)
    parts.append(model_file)
    parts.extend(data_files)
    parts.extend(bumps_opts)
    #if store is not None:
    #    parts.append("--store=" + store)
    command = " ".join(parts)

    print(command)
    os.system(command)

def run_task(task_id):
    opts = split_args()

    # Set environment put compiled sasmodels in user-specific temporary cache
    # We need this because users don't have a home directory on the individual
    # cluster nodes.
    assert opts['slurm'][0].startswith('--source_root=')
    SRC = opts['slurm'][0][14:]
    PACKAGES = ("periodictable", "sasview/src", "bumps", "sasmodels")
    os.environ['PYTHONPATH'] = ":".join(SRC+"/"+v for v in PACKAGES)
    TMP = tempfile.gettempdir()
    cache_path = os.path.join(TMP, os.environ['USER'], '.cache')
    os.environ['SAS_DLL_PATH'] = cache_path
    os.environ['XDG_CACHE_HOME'] = cache_path

    #task_store = "%s/%02d"%(opts['store'], task_id)
    task_store = "%02d"%task_id
    parts = [
       opts['python'], os.path.join(SRC, "bumps", "run.py"), "--batch",
       "--view=log",
       opts['model_file'],
       opts['data_files'][task_id-1],
       ]
    parts.extend(opts['bumps'])
    parts.append('--store='+task_store)
    if opts['resume'] is not None:
        parts.append('--resume='+os.path.join(opts['resume'], task_store))
    command = " ".join(parts)
    print(os.getcwd() + "$ " + command)
    os.system(command)


task_id = int(os.environ.get('SLURM_ARRAY_TASK_ID', -1))
if task_id == -1:
    submit_job()
else:
    run_task(task_id)




# rectangular_prism model
# Note: model title and parameter table are inserted automatically
r"""

This model provides the form factor, $P(q)$, for a rectangular prism.

Note that this model is almost totally equivalent to the existing
:ref:`parallelepiped` model.
The only difference is that the way the relevant
parameters are defined here ($a$, $b/a$, $c/a$ instead of $a$, $b$, $c$)
which allows use of polydispersity with this model while keeping the shape of
the prism (e.g. setting $b/a = 1$ and $c/a = 1$ and applying polydispersity
to *a* will generate a distribution of cubes of different sizes).


Definition
----------

The 1D scattering intensity for this model was calculated by (Mittelbach and
Porod, 1961 [#Mittelbach1961]_), but the implementation here is closer to
the equations given by (Nayuk and Huber, 2012 [#Nayuk2012]_).
Note also that the angle definitions used in the code and the present
documentation correspond to those used in Nayuk and Huber (see Fig. 1 of
that reference), with $\theta$ corresponding to $\alpha$ in that paper,
and not to the usual convention used for example in the
:ref:`parallelepiped` model.

In this model the scattering from a massive parallelepiped with an
orientation with respect to the scattering vector given by $\theta$
and $\phi$

.. math::

    A_P\,(q) =
      \frac{\sin \left( \tfrac{1}{2}qC \cos\theta \right) }{\tfrac{1}{2} qC \cos\theta}
      \,\times\,
      \frac{\sin \left( \tfrac{1}{2}qA \cos\theta \right) }{\tfrac{1}{2} qA \cos\theta}
      \,\times\ ,
      \frac{\sin \left( \tfrac{1}{2}qB \cos\theta \right) }{\tfrac{1}{2} qB \cos\theta}

where $A$, $B$ and $C$ are the sides of the parallelepiped and must fulfill
$A \le B \le C$, $\theta$ is the angle between the $z$ axis and the
$C$ axis of the parallelepiped, and $\phi$ is the angle between the
scattering vector (lying in the $xy$ plane) and the $y$ axis.

The normalized form factor in 1D is obtained averaging over all possible
orientations

.. math::
    P(q) =  \frac{2}{\pi} \int_0^{\frac{\pi}{2}} \,
    \int_0^{\frac{\pi}{2}} A_P^2(q) \, \sin\theta \, d\theta \, d\phi

And the 1D scattering intensity is calculated as

.. math::
    I(q) = \text{scale} \times V \times (\rho_\text{p} -
    \rho_\text{solvent})^2 \times P(q)

where $V$ is the volume of the rectangular prism, $\rho_\text{p}$
is the scattering length of the parallelepiped, $\rho_\text{solvent}$
is the scattering length of the solvent, and (if the data are in absolute
units) *scale* represents the volume fraction (which is unitless).

For 2d data the orientation of the particle is required, described using
angles $\theta$, $\phi$ and $\Psi$ as in the diagrams below, for further details
of the calculation and angular dispersions see :ref:`orientation`.
The angle $\Psi$ is the rotational angle around the long *C* axis. For example,
$\Psi = 0$ when the *B* axis is parallel to the *x*-axis of the detector.

For 2d, constraints must be applied during fitting to ensure that the
inequality $A < B < C$ is not violated, and hence the correct definition of
angles is preserved. The calculation will not report an error, but the
results may be not correct.

.. figure:: img/parallelepiped_angle_definition.png

    Definition of the angles for oriented core-shell parallelepipeds. Note
    that rotation $\theta$, initially in the $xz$ plane, is carried out
    first, then rotation $\phi$ about the $z$ axis, finally rotation $\Psi$
    is now around the axis of the cylinder. The neutron or X-ray beam is
    along the $z$ axis.

.. figure:: img/parallelepiped_angle_projection.png

    Examples of the angles for oriented rectangular prisms against the
    detector plane.


Validation
----------

Validation of the code was conducted by comparing the output of the 1D model
to the output of the existing :ref:`parallelepiped` model.


References
----------

See also Onsager [#Onsager1949]_.

.. [#Mittelbach1961] P Mittelbach and G Porod, *Acta Physica Austriaca*,
   14 (1961) 185-211
.. [#Nayuk2012] R Nayuk and K Huber, *Z. Phys. Chem.*, 226 (2012) 837-854
.. [#Onsager1949] L Onsager, *Ann. New York Acad. Sci.*, 51 (1949) 627-659

Authorship and Verification
----------------------------

* **Author:**
* **Last Modified by:**
* **Last Reviewed by:**
"""

import numpy as np
from numpy import inf

name = "rectangular_prism"
title = "Rectangular parallelepiped with uniform scattering length density."
description = """
    I(q)= scale*V*(sld - sld_solvent)^2*P(q,theta,phi)+background
        P(q,theta,phi) = (2/pi) * double integral from 0 to pi/2 of ...
           AP^2(q)*sin(theta)*dtheta*dphi
        AP = S(q*C*cos(theta)/2) * S(q*A*sin(theta)*sin(phi)/2) * S(q*B*sin(theta)*cos(phi)/2)
        S(x) = sin(x)/x
"""
category = "shape:parallelepiped"

#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["sld", "1e-6/Ang^2", 6.3, [-inf, inf], "sld",
               "Parallelepiped scattering length density"],
              ["sld_solvent", "1e-6/Ang^2", 1, [-inf, inf], "sld",
               "Solvent scattering length density"],
              ["length_a", "Ang", 35, [0, inf], "volume",
               "Shorter side of the parallelepiped"],
              ["b2a_ratio", "", 1, [0, inf], "volume",
               "Ratio sides b/a"],
              ["c2a_ratio", "", 1, [0, inf], "volume",
               "Ratio sides c/a"],
              ["theta", "degrees", 0, [-360, 360], "orientation",
               "c axis to beam angle"],
              ["phi", "degrees", 0, [-360, 360], "orientation",
               "rotation about beam"],
              ["psi", "degrees", 0, [-360, 360], "orientation",
               "rotation about c axis"],
             ]

source = ["lib/gauss76.c", "rectangular_prism.c"]
have_Fq = True
radius_effective_modes = [
    "equivalent cylinder excluded volume", "equivalent volume sphere",
    "half length_a", "half length_b", "half length_c",
    "equivalent circular cross-section", "half ab diagonal", "half diagonal",
    ]

def random():
    """Return a random parameter set for the model."""
    a, b, c = 10**np.random.uniform(1, 4.7, size=3)
    pars = dict(
        length_a=a,
        b2a_ratio=b/a,
        c2a_ratio=c/a,
    )
    return pars

tests = [[{}, 0.2, 0.375248406825],
         [{}, [0.2], [0.375248406825]],
        ]



"""
SAS data representations.

Plotting functions for data sets:

    :func:`plot_data` plots the data file.

    :func:`plot_theory` plots a calculated result from the model.

Wrappers for the sasview data loader and data manipulations:

    :func:`load_data` loads a sasview data file.

    :func:`set_beam_stop` masks the beam stop from the data.

    :func:`set_half` selects the right or left half of the data, which can
    be useful for shear measurements which have not been properly corrected
    for path length and reflections.

    :func:`set_top` cuts the top part off the data.


Empty data sets for evaluating models without data:

    :func:`empty_data1D` creates an empty dataset, which is useful for plotting
    a theory function before the data is measured.

    :func:`empty_data2D` creates an empty 2D dataset.

Note that the empty datasets use a minimal representation of the SasView
objects so that models can be run without SasView on the path.  You could
also use these for your own data loader.

"""
import traceback
from functools import wraps

import numpy as np  # type: ignore
from numpy import sqrt, sin, cos, pi

# pylint: disable=unused-import
try:
    from typing import Union, Dict, List, Optional, Tuple, Callable
    Data = Union["Data1D", "Data2D", "SesansData"]
    OptArray = Optional[np.ndarray]
    OptLimits = Optional[Tuple[float, float]]
    OptString = Optional[str]
except ImportError:
    pass
# pylint: enable=unused-import

def load_data(filename, index=0):
    # type: (str, int) -> Data
    """
    Load data using a sasview loader.
    """
    try:
        from sasdata.dataloader.loader import Loader  # type: ignore
    except ImportError as ie:
        raise ImportError(f"{ie.name} is not available. Add sasdata to the python path.")
    loader = Loader()
    # Allow for one part in multipart file
    if '[' in filename:
        filename, indexstr = filename[:-1].split('[')
        index = int(indexstr)
    datasets = loader.load(filename)
    if not datasets:  # None or []
        raise IOError("Data %r could not be loaded" % filename)
    if not isinstance(datasets, list):
        datasets = [datasets]
    for data in datasets:
        if getattr(data, 'isSesans', False):
            pass
        elif hasattr(data, 'x'):
            data.qmin, data.qmax = data.x.min(), data.x.max()
            data.mask = (np.isnan(data.y) if data.y is not None
                         else np.zeros_like(data.x, dtype='bool'))
        elif hasattr(data, 'qx_data'):
            data.mask = ~data.mask
    return datasets[index] if index != 'all' else datasets


def set_beam_stop(data, radius, outer=None):
    # type: (Data, float, Optional[float]) -> None
    """
    Add a beam stop of the given *radius*.  If *outer*, make an annulus.
    """
    try:
        from sasdata.data_util.manipulations import Ringcut
    except ImportError as ie:
        raise ImportError(f"{ie.name} is not available. Add sasdata to the python path.")
    if hasattr(data, 'qx_data'):
        data.mask = Ringcut(0, radius)(data)
        if outer is not None:
            data.mask += Ringcut(outer, np.inf)(data)
    else:
        data.mask = (data.x < radius)
        if outer is not None:
            data.mask |= (data.x >= outer)


def set_half(data, half):
    # type: (Data, str) -> None
    """
    Select half of the data, either "right" or "left".
    """
    try:
        from sasdata.data_util.manipulations import Boxcut
    except ImportError as ie:
        raise ImportError(f"{ie.name} is not available. Add sasdata to the python path.")
    if half == 'right':
        data.mask += \
            Boxcut(x_min=-np.inf, x_max=0.0, y_min=-np.inf, y_max=np.inf)(data)
    if half == 'left':
        data.mask += \
            Boxcut(x_min=0.0, x_max=np.inf, y_min=-np.inf, y_max=np.inf)(data)


def set_top(data, cutoff):
    # type: (Data, float) -> None
    """
    Chop the top off the data, above *cutoff*.
    """
    try:
        from sasdata.data_util.manipulations import Boxcut
    except ImportError as ie:
        raise ImportError(f"{ie.name} is not available. Add sasdata to the python path.")
    data.mask += \
        Boxcut(x_min=-np.inf, x_max=np.inf, y_min=-np.inf, y_max=cutoff)(data)


class Source:
    ...
class Sample:
    ...

def _as_numpy(data):
    return None if data is None else np.asarray(data)

class Data1D(object):
    """
    1D data object.

    Note that this definition matches the attributes from sasview, with
    some generic 1D data vectors and some SAS specific definitions.  Some
    refactoring to allow consistent naming conventions between 1D, 2D and
    SESANS data would be helpful.

    **Attributes**

    *x*, *dx*: $q$ vector and gaussian resolution

    *y*, *dy*: $I(q)$ vector and measurement uncertainty

    *mask*: values to include in plotting/analysis

    *dxl*: slit widths for slit smeared data, with *dx* ignored

    *qmin*, *qmax*: range of $q$ values in *x*

    *filename*: label for the data line

    *_xaxis*, *_xunit*: label and units for the *x* axis

    *_yaxis*, *_yunit*: label and units for the *y* axis
    """
    def __init__(self, x=None, y=None, dx=None, dy=None):
        # type: (OptArray, OptArray, OptArray, OptArray) -> None
        self.x, self.dx = _as_numpy(x), _as_numpy(dx)
        self.y, self.dy = _as_numpy(y), _as_numpy(dy)
        self.dxl = None
        self.filename = None
        self.qmin = self.x.min() if self.x is not None else np.nan
        self.qmax = self.x.max() if self.x is not None else np.nan
        # TODO: why is 1D mask False and 2D mask True?
        self.mask = (np.isnan(y) if y is not None
                     else np.zeros_like(x, 'b') if x is not None
                     else None)
        self._xaxis, self._xunit = "x", ""
        self._yaxis, self._yunit = "y", ""

    def xaxis(self, label, unit):
        # type: (str, str) -> None
        """
        set the x axis label and unit
        """
        self._xaxis = label
        self._xunit = unit

    def yaxis(self, label, unit):
        # type: (str, str) -> None
        """
        set the y axis label and unit
        """
        self._yaxis = label
        self._yunit = unit

class SesansData(Data1D):
    """
    SESANS data object.

    This is just :class:`Data1D` with a wavelength parameter.

    *x* is spin echo length and *y* is polarization (P/P0).
    """
    isSesans = True
    def __init__(self, **kw):
        Data1D.__init__(self, **kw)
        self.lam = None  # type: OptArray
        self.xaxis("SE length", "A")
        self.yaxis("log(P)/(t L^2)", "1/A^2 1/cm")

class Data2D(object):
    """
    2D data object.

    Note that this definition matches the attributes from sasview. Some
    refactoring to allow consistent naming conventions between 1D, 2D and
    SESANS data would be helpful.

    **Attributes**

    *qx_data*, *dqx_data*: $q_x$ matrix and gaussian resolution

    *qy_data*, *dqy_data*: $q_y$ matrix and gaussian resolution

    *data*, *err_data*: $I(q)$ matrix and measurement uncertainty

    *mask*: values to exclude from plotting/analysis

    *qmin*, *qmax*: range of $q$ values in *x*

    *filename*: label for the data line

    *_xaxis*, *_xunit*: label and units for the *x* axis

    *_yaxis*, *_yunit*: label and units for the *y* axis

    *_zaxis*, *_zunit*: label and units for the *y* axis

    *Q_unit*, *I_unit*: units for Q and intensity

    *x_bins*, *y_bins*: grid steps in *x* and *y* directions
    """
    def __init__(self, x=None, y=None, z=None, dx=None, dy=None, dz=None):
        # type: (OptArray, OptArray, OptArray, OptArray, OptArray, OptArray) -> None
        self.qx_data, self.dqx_data = _as_numpy(x), _as_numpy(dx)
        self.qy_data, self.dqy_data = _as_numpy(y), _as_numpy(dy)
        self.data, self.err_data = _as_numpy(z), _as_numpy(dz)
        self.mask = (np.isnan(z) if z is not None
                     else np.zeros_like(x, dtype='bool') if x is not None
                     else None)
        self.q_data = np.sqrt(self.qx_data**2 + self.qy_data**2)
        self.qmin = 1e-16
        self.qmax = np.inf
        self.detector = []
        self.source = Source()
        self.Q_unit = "1/A"
        self.I_unit = "1/cm"
        self.xaxis("Q_x", "1/A")
        self.yaxis("Q_y", "1/A")
        self.zaxis("Intensity", "1/cm")
        self._xaxis, self._xunit = "x", ""
        self._yaxis, self._yunit = "y", ""
        self._zaxis, self._zunit = "z", ""
        self.x_bins, self.y_bins = None, None
        self.filename = None

    def xaxis(self, label, unit):
        # type: (str, str) -> None
        """
        set the x axis label and unit
        """
        self._xaxis = label
        self._xunit = unit

    def yaxis(self, label, unit):
        # type: (str, str) -> None
        """
        set the y axis label and unit
        """
        self._yaxis = label
        self._yunit = unit

    def zaxis(self, label, unit):
        # type: (str, str) -> None
        """
        set the y axis label and unit
        """
        self._zaxis = label
        self._zunit = unit


class Vector(object):
    """
    3-space vector of *x*, *y*, *z*
    """
    def __init__(self, x=None, y=None, z=None):
        # type: (float, float, Optional[float]) -> None
        self.x, self.y, self.z = x, y, z

class Detector(object):
    """
    Detector attributes.
    """
    def __init__(self, pixel_size=(None, None), distance=None):
        # type: (Tuple[float, float], float) -> None
        self.pixel_size = Vector(*pixel_size)
        self.distance = distance

class Source(object):
    """
    Beam attributes.
    """
    def __init__(self):
        # type: () -> None
        self.wavelength = np.nan
        self.wavelength_unit = "A"

class Sample(object):
    """
    Sample attributes.
    """
    def __init__(self):
        # type: () -> None
        pass

def empty_sesans(z, wavelength=None, zacceptance=None):
    data = SesansData(x=z, y=None, dx=None, dy=None)
    data.filename = "fake data"
    DEFAULT_WAVELENGTH = 5
    if wavelength is None:
        wavelength = DEFAULT_WAVELENGTH
    if np.isscalar(wavelength):
        wavelength = np.full_like(z, wavelength)
    if zacceptance is None:
        zacceptance = (np.pi/2, 'radians')
    source = Source()
    source.wavelength = wavelength
    source.wavelength_unit = "A"
    sample = Sample()
    sample.zacceptance = zacceptance
    data.source = source
    data.sample = sample
    return data

def empty_data1D(q, resolution=0.0, L=0., dL=0.):
    # type: (np.ndarray, float, float, float) -> Data1D
    r"""
    Create empty 1D data using the given *q* as the x value.

    rms *resolution* $\Delta q/q$ defaults to 0%.  If wavelength *L* and rms
    wavelength divergence *dL* are defined, then *resolution* defines
    rms $\Delta \theta/\theta$ for the lowest *q*, with $\theta$ derived from
    $q = 4\pi/\lambda \sin(\theta)$.
    """

    #Iq = 100 * np.ones_like(q)
    #dIq = np.sqrt(Iq)
    Iq, dIq = None, None
    q = np.asarray(q)
    if L != 0 and resolution != 0:
        theta = np.arcsin(q*L/(4*pi))
        dtheta = theta[0]*resolution
        ## Solving Gaussian error propagation from
        ##   Dq^2 = (dq/dL)^2 DL^2 + (dq/dtheta)^2 Dtheta^2
        ## gives
        ##   (Dq/q)^2 = (DL/L)**2 + (Dtheta/tan(theta))**2
        ## Take the square root and multiply by q, giving
        ##   Dq = (4*pi/L) * sqrt((sin(theta)*dL/L)**2 + (cos(theta)*dtheta)**2)
        dq = (4*pi/L) * sqrt((sin(theta)*dL/L)**2 + (cos(theta)*dtheta)**2)
    else:
        dq = resolution * q
    data = Data1D(q, Iq, dx=dq, dy=dIq)
    data.filename = "fake data"
    return data


def empty_data2D(qx, qy=None, resolution=0.0):
    # type: (np.ndarray, OptArray, float) -> Data2D
    """
    Create empty 2D data using the given mesh.

    If *qy* is missing, create a square mesh with *qy=qx*.

    *resolution* dq/q defaults to 5%.
    """
    if qy is None:
        qy = qx
    qx, qy = np.asarray(qx), np.asarray(qy)
    # 5% dQ/Q resolution
    Qx, Qy = np.meshgrid(qx, qy)
    Qx, Qy = Qx.flatten(), Qy.flatten()
    Iq = 100 * np.ones_like(Qx)  # type: np.ndarray
    dIq = np.sqrt(Iq)
    if resolution != 0:
        # https://www.ncnr.nist.gov/staff/hammouda/distance_learning/chapter_15.pdf
        # Should have an additional constant which depends on distances and
        # radii of the aperture, pixel dimensions and wavelength spread
        # Instead, assume radial dQ/Q is constant, and perpendicular matches
        # radial (which instead it should be inverse).
        Q = np.sqrt(Qx**2 + Qy**2)
        dqx = resolution * Q
        dqy = resolution * Q
    else:
        dqx = dqy = None

    data = Data2D(x=Qx, y=Qy, z=Iq, dx=dqx, dy=dqy, dz=dIq)
    data.x_bins = qx
    data.y_bins = qy
    data.filename = "fake data"

    # pixel_size in mm, distance in m
    detector = Detector(pixel_size=(5, 5), distance=4)
    data.detector.append(detector)
    data.source.wavelength = 5 # angstroms
    data.source.wavelength_unit = "A"
    return data


def plot_data(data, view=None, limits=None):
    # type: (Data, str, OptLimits) -> None
    """
    Plot data loaded by the sasview loader.

    *data* is a sasview data object, either 1D, 2D or SESANS.

    *view* is log, linear or normed.

    *limits* sets the intensity limits on the plot; if None then the limits
    are inferred from the data.
    """
    # Note: kind of weird using the plot result functions to plot just the
    # data, but they already handle the masking and graph markup already, so
    # do not repeat.
    if hasattr(data, 'isSesans') and data.isSesans:
        _plot_result_sesans(data, None, None, view, use_data=True, limits=limits)
    elif hasattr(data, 'qx_data') and not getattr(data, 'radial', False):
        _plot_result2D(data, None, None, view, use_data=True, limits=limits)
    else:
        _plot_result1D(data, None, None, view, use_data=True, limits=limits)


def plot_theory(data, theory, resid=None, view=None, use_data=True,
                limits=None, Iq_calc=None):
    # type: (Data, OptArray, OptArray, OptString, bool, OptLimits, OptArray) -> None
    """
    Plot theory calculation.

    *data* is needed to define the graph properties such as labels and
    units, and to define the data mask.

    *theory* is a matrix of the same shape as the data.

    *view* is log, linear or normed

    *use_data* is True if the data should be plotted as well as the theory.

    *limits* sets the intensity limits on the plot; if None then the limits
    are inferred from the data. If (-inf, inf) then use auto limits.

    *Iq_calc* is the raw theory values without resolution smearing
    """
    if limits is not None and np.isinf(limits[0]):
        limits = None
    if hasattr(data, 'isSesans') and data.isSesans:
        _plot_result_sesans(data, theory, resid, view, use_data=True, limits=limits)
    elif hasattr(data, 'qx_data') and not getattr(data, 'radial', False):
        _plot_result2D(data, theory, resid, view, use_data, limits=limits)
    else:
        _plot_result1D(data, theory, resid, view, use_data,
                       limits=limits, Iq_calc=Iq_calc)


def protect(func):
    # type: (Callable) -> Callable
    """
    Decorator to wrap calls in an exception trapper which prints the
    exception and continues.  Keyboard interrupts are ignored.
    """
    @wraps(func)
    def wrapper(*args, **kw):
        """
        Trap and print errors from function.
        """
        try:
            return func(*args, **kw)
        except Exception as exc:
            print("Traceback (most recent call last):")
            print("".join(traceback.format_list(traceback.extract_stack(limit=4)[:-1]))[:-1])
            print(f"{exc.__class__.__name__}: {exc}")

    return wrapper


@protect
def _plot_result1D(data, theory, resid, view, use_data,
                   limits=None, Iq_calc=None):
    # type: (Data1D, OptArray, OptArray, str, bool, OptLimits, OptArray) -> None
    """
    Plot the data and residuals for 1D data.
    """
    import matplotlib.pyplot as plt  # type: ignore
    from numpy.ma import masked_array, masked  # type: ignore

    # Default to 'log' view
    if view is None:
        view = 'log'

    if getattr(data, 'radial', False):
        data.x = data.q_data
        data.y = data.data

    use_data = use_data and data.y is not None
    use_theory = theory is not None
    use_resid = resid is not None
    use_calc = use_theory and Iq_calc is not None
    num_plots = (use_data or use_theory) + use_calc + use_resid
    non_positive_x = (data.x <= 0.0).any()

    scale = 1e8 * data.x**4 if view == 'q4' else 1.0

    if use_data or use_theory:
        if num_plots > 1:
            plt.subplot(1, num_plots, 1)

        #print(vmin, vmax)
        all_positive = True
        some_present = False
        if use_data:
            mdata = masked_array(data.y, data.mask.copy())
            mdata[~np.isfinite(mdata)] = masked
            if view == 'log':
                mdata[mdata <= 0] = masked
            plt.errorbar(data.x, scale*mdata, yerr=scale*data.dy, fmt='.')
            all_positive = all_positive and (mdata > 0).all()
            some_present = some_present or (mdata.count() > 0)


        if use_theory:
            # Theory values are only calculated where the data is not masked,
            # so restrict data.x and scale to only those points.
            # Note: masks merge, so any masked theory points will stay masked,
            # and the data mask will be added to it.
            #mtheory = masked_array(theory, data.mask.copy())
            theory_x = data.x[data.mask == 0]
            theory_scale = scale if np.isscalar(scale) else scale[data.mask == 0]
            mtheory = masked_array(theory)
            mtheory[~np.isfinite(mtheory)] = masked
            if view == 'log':
                mtheory[mtheory <= 0] = masked
            plt.plot(theory_x, theory_scale*mtheory, '-')
            all_positive = all_positive and (mtheory > 0).all()
            some_present = some_present or (mtheory.count() > 0)

        if limits is not None:
            plt.ylim(*limits)


        plt.xscale('linear' if not some_present or non_positive_x else 'log')
        plt.yscale('log' if some_present and view == 'log' else 'linear')
        plt.xlabel("$q$/A$^{-1}$")
        plt.ylabel('$10^8 q^4 I(q)$' if view == 'q4' else '$I(q)$')
        title = ("data and model" if use_theory and use_data
                 else "data" if use_data
                 else "model")
        plt.title(title)

    if use_calc:
        # Only have use_calc if have use_theory
        plt.subplot(1, num_plots, 2)
        qx, qy, Iqxy = Iq_calc
        plt.pcolormesh(qx, qy[qy > 0], np.log10(Iqxy[qy > 0, :]))
        plt.xlabel("$q_x$/A$^{-1}$")
        plt.xlabel("$q_y$/A$^{-1}$")
        plt.xscale('log')
        plt.yscale('log')
        #plt.axis('equal')

    if use_resid:
        theory_x = data.x[data.mask == 0]
        mresid = masked_array(resid)
        mresid[~np.isfinite(mresid)] = masked
        some_present = (mresid.count() > 0)

        if num_plots > 1:
            plt.subplot(1, num_plots, use_calc + 2)
        plt.plot(theory_x, mresid, '.')
        plt.xlabel("$q$/A$^{-1}$")
        plt.ylabel('residuals')
        plt.title('(model - Iq)/dIq')
        plt.xscale('linear' if not some_present or non_positive_x else 'log')
        plt.yscale('linear')


@protect
def _plot_result_sesans(data, theory, resid, view, use_data, limits=None):
    # type: (SesansData, OptArray, OptArray, OptString, bool, OptLimits) -> None
    """
    Plot SESANS results.
    """
    import matplotlib.pyplot as plt  # type: ignore
    use_data = use_data and data.y is not None
    use_theory = theory is not None
    use_resid = resid is not None
    num_plots = (use_data or use_theory) + use_resid

    normed = (view == "normed")
    #normed = True
    offset, scale = 0, 1
    if normed and theory is not None:
        offset, scale = theory[-1], theory[0] - theory[-1]

    if use_data or use_theory:
        is_tof = data.lam is not None and (data.lam != data.lam[0]).any()
        if num_plots > 1:
            plt.subplot(1, num_plots, 1)
        if use_data:
            if is_tof:
                plt.errorbar(data.x, np.log(data.y)/(data.lam*data.lam),
                             yerr=data.dy/data.y/(data.lam*data.lam))
            else:
                #plt.errorbar(data.x, data.y, yerr=data.dy)
                plt.errorbar(data.x, (data.y-offset)/scale, yerr=data.dy/scale)
        if theory is not None:
            if is_tof:
                plt.plot(data.x, np.log(theory)/(data.lam*data.lam), '-')
            else:
                #plt.plot(data.x, theory, '-')
                plt.plot(data.x, (theory-offset)/scale, '-')
        if limits is not None:
            plt.ylim(*limits)

        plt.xlabel('spin echo length ({})'.format(data._xunit))
        plt.ylabel(r'$\log(P)/(t\lambda^2) (\mathrm{A}^{-2}\mathrm{cm}^{-1})$')
        plt.xscale('log')


    if resid is not None:
        if num_plots > 1:
            plt.subplot(1, num_plots, (use_data or use_theory) + 1)
        plt.plot(data.x, resid, 'x')
        plt.xlabel('spin echo length ({})'.format(data._xunit))
        plt.ylabel('polarization residuals')
        plt.xscale('log')


@protect
def _plot_result2D(data, theory, resid, view, use_data, limits=None):
    # type: (Data2D, OptArray, OptArray, str, bool, OptLimits) -> None
    """
    Plot the data and residuals for 2D data.
    """
    import matplotlib.pyplot as plt  # type: ignore
    if view is None:
        view = 'log'
    use_data = use_data and data.data is not None
    use_theory = theory is not None
    use_resid = resid is not None
    num_plots = use_data + use_theory + use_resid

    # Put theory and data on a common colormap scale
    vmin, vmax = np.inf, -np.inf
    target = None  # type: OptArray
    if use_data:
        target = data.data[~data.mask]
        datamin = target[target > 0].min() if view == 'log' else target.min()
        datamax = target.max()
        vmin = min(vmin, datamin)
        vmax = max(vmax, datamax)
    if use_theory:
        theorymin = theory[theory > 0].min() if view == 'log' else theory.min()
        theorymax = theory.max()
        vmin = min(vmin, theorymin)
        vmax = max(vmax, theorymax)

    # Override data limits from the caller
    if limits is not None:
        vmin, vmax = limits

    # Plot data
    if use_data:
        if num_plots > 1:
            plt.subplot(1, num_plots, 1)
        _plot_2d_signal(data, target, view=view, vmin=vmin, vmax=vmax)
        plt.title('data')
        h = plt.colorbar()
        h.set_label('$I(q)$')

    # plot theory
    if use_theory:
        if num_plots > 1:
            plt.subplot(1, num_plots, use_data+1)
        _plot_2d_signal(data, theory, view=view, vmin=vmin, vmax=vmax)
        plt.title('theory')
        h = plt.colorbar()
        h.set_label(r'$\log_{10}I(q)$' if view == 'log'
                    else r'$q^4 I(q)$' if view == 'q4'
                    else '$I(q)$')

    # plot resid
    if use_resid:
        if num_plots > 1:
            plt.subplot(1, num_plots, use_data+use_theory+1)
        _plot_2d_signal(data, resid, view='linear')
        plt.title('residuals')
        h = plt.colorbar()
        h.set_label(r'$\Delta I(q)$')


@protect
def _plot_2d_signal(data, signal, vmin=None, vmax=None, view=None):
    # type: (Data2D, np.ndarray, Optional[float], Optional[float], str) -> Tuple[float, float]
    """
    Plot the target value for the data.  This could be the data itself,
    the theory calculation, or the residuals.

    *scale* can be 'log' for log scale data, or 'linear'.
    """
    import matplotlib.pyplot as plt  # type: ignore
    from numpy.ma import masked_array  # type: ignore

    if view is None:
        view = 'log'

    image = np.zeros_like(data.qx_data)
    image[~data.mask] = signal
    valid = np.isfinite(image) & ~data.mask
    if view == 'log':
        valid &= image > 0
        if vmin is None:
            vmin = image[valid].min()
        if vmax is None:
            vmax = image[valid].max()
        image[valid] = np.log10(image[valid])
    elif view == 'q4':
        image[valid] *= (data.qx_data[valid]**2+data.qy_data[valid]**2)**2
        if vmin is None:
            vmin = image[valid].min()
        if vmax is None:
            vmax = image[valid].max()
    else:
        if vmin is None:
            vmin = image[valid].min()
        if vmax is None:
            vmax = image[valid].max()

    image[~valid] = 0
    #plottable = Iq
    plottable = masked_array(image, ~valid)
    # Divide range by 10 to convert from angstroms to nanometers
    xmin, xmax = min(data.qx_data), max(data.qx_data)
    ymin, ymax = min(data.qy_data), max(data.qy_data)
    if view == 'log':
        vmin_scaled, vmax_scaled = np.log10(vmin), np.log10(vmax)
    else:
        vmin_scaled, vmax_scaled = vmin, vmax
    #nx, ny = len(data.x_bins), len(data.y_bins)
    _, _, image = _build_matrix(data, plottable)
    plt.imshow(image,
               interpolation='nearest', aspect=1, origin='lower',
               extent=[xmin, xmax, ymin, ymax],
               vmin=vmin_scaled, vmax=vmax_scaled)
    plt.xlabel("$q_x$/A$^{-1}$")
    plt.ylabel("$q_y$/A$^{-1}$")
    return vmin, vmax


# === The following is modified from sas.sasgui.plottools.PlotPanel
def _build_matrix(self, plottable):
    """
    Build a matrix for 2d plot from a vector
    Returns a matrix (image) with ~ square binning
    Requirement: need 1d array formats of
    self.data, self.qx_data, and self.qy_data
    where each one corresponds to z, x, or y axis values

    """
    # No qx or qy given in a vector format
    if (self.qx_data is None or self.qy_data is None
            or self.qx_data.ndim != 1 or self.qy_data.ndim != 1):
        return self.x_bins, self.y_bins, plottable

    # maximum # of loops to fillup_pixels
    # otherwise, loop could never stop depending on data
    max_loop = 1
    # get the x and y_bin arrays.
    x_bins, y_bins = _get_bins(self)
    # set zero to None

    #Note: Can not use scipy.interpolate.Rbf:
    # 'cause too many data points (>10000)<=JHC.
    # 1d array to use for weighting the data point averaging
    #when they fall into a same bin.
    weights_data = np.ones([self.data.size])
    # get histogram of ones w/len(data); this will provide
    #the weights of data on each bins
    weights, _, _ = np.histogram2d(
        x=self.qy_data, y=self.qx_data, bins=[y_bins, x_bins],
        weights=weights_data)
    # get histogram of data, all points into a bin in a way of summing
    image, _, _ = np.histogram2d(
        x=self.qy_data, y=self.qx_data, bins=[y_bins, x_bins],
        weights=plottable)
    # Now, normalize the image by weights only for weights>1:
    # If weight == 1, there is only one data point in the bin so
    # that no normalization is required.
    image[weights > 1] = image[weights > 1] / weights[weights > 1]
    # Set image bins w/o a data point (weight==0) as None (was set to zero
    # by histogram2d.)
    image[weights == 0] = None

    # Fill empty bins with 8 nearest neighbors only when at least
    #one None point exists
    loop = 0

    # do while loop until all vacant bins are filled up up
    #to loop = max_loop
    while (weights == 0).any():
        if loop >= max_loop:  # this protects never-ending loop
            break
        image = _fillup_pixels(image=image, weights=weights)
        loop += 1

    return x_bins, y_bins, image

def _get_bins(self):
    """
    get bins
    set x_bins and y_bins into self, 1d arrays of the index with
    ~ square binning
    Requirement: need 1d array formats of
    self.qx_data, and self.qy_data
    where each one corresponds to  x, or y axis values
    """
    # find max and min values of qx and qy
    xmax = self.qx_data.max()
    xmin = self.qx_data.min()
    ymax = self.qy_data.max()
    ymin = self.qy_data.min()

    # calculate the range of qx and qy: this way, it is a little
    # more independent
    x_size = xmax - xmin
    y_size = ymax - ymin

    # estimate the # of pixels on each axes
    npix_y = int(np.floor(np.sqrt(len(self.qy_data))))
    npix_x = int(np.floor(len(self.qy_data) / npix_y))

    # bin size: x- & y-directions
    xstep = x_size / (npix_x - 1)
    ystep = y_size / (npix_y - 1)

    # max and min taking account of the bin sizes
    xmax = xmax + xstep / 2.0
    xmin = xmin - xstep / 2.0
    ymax = ymax + ystep / 2.0
    ymin = ymin - ystep / 2.0

    # store x and y bin centers in q space
    x_bins = np.linspace(xmin, xmax, npix_x)
    y_bins = np.linspace(ymin, ymax, npix_y)

    return x_bins, y_bins

def _fillup_pixels(image=None, weights=None):
    """
    Fill z values of the empty cells of 2d image matrix
    with the average over up-to next nearest neighbor points

    :param image: (2d matrix with some zi = None)

    :return: image (2d array )

    :TODO: Find better way to do for-loop below
    """
    # No image matrix given
    if (image is None or np.ndim(image) != 2
            or np.isfinite(image).all()
            or weights is None):
        return image
    # Get bin size in y and x directions
    len_x, len_y = image.shape[1], image.shape[0]
    temp_image = np.zeros([len_y, len_x])
    weit = np.zeros([len_y, len_x])
    # do for-loop for all pixels
    for n_y in range(len_y):
        for n_x in range(len_x):
            # find only null pixels
            if weights[n_y][n_x] > 0 or np.isfinite(image[n_y][n_x]):
                continue
            # find 4 nearest neighbors
            # check where or not it is at the corner
            if n_y != 0 and np.isfinite(image[n_y - 1][n_x]):
                temp_image[n_y][n_x] += image[n_y - 1][n_x]
                weit[n_y][n_x] += 1
            if n_x != 0 and np.isfinite(image[n_y][n_x - 1]):
                temp_image[n_y][n_x] += image[n_y][n_x - 1]
                weit[n_y][n_x] += 1
            if n_y != len_y - 1 and np.isfinite(image[n_y + 1][n_x]):
                temp_image[n_y][n_x] += image[n_y + 1][n_x]
                weit[n_y][n_x] += 1
            if n_x != len_x - 1 and np.isfinite(image[n_y][n_x + 1]):
                temp_image[n_y][n_x] += image[n_y][n_x + 1]
                weit[n_y][n_x] += 1
            # go 4 next nearest neighbors when no non-zero
            # neighbor exists
            if (n_y != 0 and n_x != 0
                    and np.isfinite(image[n_y - 1][n_x - 1])):
                temp_image[n_y][n_x] += image[n_y - 1][n_x - 1]
                weit[n_y][n_x] += 1
            if (n_y != len_y - 1 and n_x != 0
                    and np.isfinite(image[n_y + 1][n_x - 1])):
                temp_image[n_y][n_x] += image[n_y + 1][n_x - 1]
                weit[n_y][n_x] += 1
            if (n_y != len_y and n_x != len_x - 1
                    and np.isfinite(image[n_y - 1][n_x + 1])):
                temp_image[n_y][n_x] += image[n_y - 1][n_x + 1]
                weit[n_y][n_x] += 1
            if (n_y != len_y - 1 and n_x != len_x - 1
                    and np.isfinite(image[n_y + 1][n_x + 1])):
                temp_image[n_y][n_x] += image[n_y + 1][n_x + 1]
                weit[n_y][n_x] += 1

    # get it normalized
    ind = (weit > 0)
    image[ind] = temp_image[ind] / weit[ind]

    return image


def demo():
    # type: () -> None
    """
    Load and plot a SAS dataset.
    """
    data = load_data('DEC07086.DAT')
    set_beam_stop(data, 0.004)
    plot_data(data)
    import matplotlib.pyplot as plt  # type: ignore
    plt.show()


if __name__ == "__main__":
    demo()



# triaxial ellipsoid model
# Note: model title and parameter table are inserted automatically
r"""
Definition
----------

.. figure:: img/triaxial_ellipsoid_geometry.jpg

    Ellipsoid with $R_a$ as *radius_equat_minor*, $R_b$ as *radius_equat_major*
    and $R_c$ as *radius_polar*.

Given an ellipsoid

.. math::

    \frac{X^2}{R_a^2} + \frac{Y^2}{R_b^2} + \frac{Z^2}{R_c^2} = 1

the scattering for randomly oriented particles is defined by the average over
all orientations $\Omega$ of:

.. math::

    P(q) = \text{scale}(\Delta\rho)^2\frac{V}{4 \pi}\int_\Omega\Phi^2(qr)\,d\Omega
           + \text{background}

where

.. math::

    \Phi(qr) &= 3 j_1(qr)/qr = 3 (\sin qr - qr \cos qr)/(qr)^3 \\
    r^2 &= R_a^2e^2 + R_b^2f^2 + R_c^2g^2 \\
    V &= \tfrac{4}{3} \pi R_a R_b R_c

The $e$, $f$ and $g$ terms are the projections of the orientation vector on $X$,
$Y$ and $Z$ respectively.  Keeping the orientation fixed at the canonical
axes, we can integrate over the incident direction using polar angle
$-\pi/2 \le \gamma \le \pi/2$ and equatorial angle $0 \le \phi \le 2\pi$
(as defined in ref [1]),

 .. math::

     \langle\Phi^2\rangle = \int_0^{2\pi} \int_{-\pi/2}^{\pi/2} \Phi^2(qr)
                                                \cos \gamma\,d\gamma d\phi

with $e = \cos\gamma \sin\phi$, $f = \cos\gamma \cos\phi$ and $g = \sin\gamma$.
A little algebra yields

.. math::

    r^2 = b^2(p_a \sin^2 \phi \cos^2 \gamma + 1 + p_c \sin^2 \gamma)

for

.. math::

    p_a = \frac{a^2}{b^2} - 1 \text{ and } p_c = \frac{c^2}{b^2} - 1

Due to symmetry, the ranges can be restricted to a single quadrant
$0 \le \gamma \le \pi/2$ and $0 \le \phi \le \pi/2$, scaling the resulting
integral by 8. The computation is done using the substitution $u = \sin\gamma$,
$du = \cos\gamma\,d\gamma$, giving

.. math::

    \langle\Phi^2\rangle &= 8 \int_0^{\pi/2} \int_0^1 \Phi^2(qr) du d\phi \\
    r^2 &= b^2(p_a \sin^2(\phi)(1 - u^2) + 1 + p_c u^2)

Though for convenience we describe the three radii of the ellipsoid as
equatorial and polar, they may be given in $any$ size order. To avoid
multiple solutions, especially with Monte-Carlo fit methods, it may be
advisable to restrict their ranges. For typical small angle diffraction
situations there may be a number of closely similar "best fits", so some
trial and error, or fixing of some radii at expected values, may help.

To provide easy access to the orientation of the triaxial ellipsoid, we
define the axis of the cylinder using the angles $\theta$, $\phi$ and $\psi$.
These angles are defined analogously to the elliptical_cylinder below, note
that angle $\phi$ is now NOT the same as in the equations above.

.. figure:: img/elliptical_cylinder_angle_definition.png

    Definition of angles for oriented triaxial ellipsoid, where radii are for
    illustration here $a < b << c$ and angle $\Psi$ is a rotation around the
    axis of the particle.

For oriented ellipsoids the *theta*, *phi* and *psi* orientation parameters
will appear when fitting 2D data, see the :ref:`elliptical-cylinder` model
for further information.

.. _triaxial-ellipsoid-angles:

.. figure:: img/triaxial_ellipsoid_angle_projection.png

    Some examples for an oriented triaxial ellipsoid.

The radius-of-gyration for this system is  $R_g^2 = (R_a R_b R_c)^2/5$.

The contrast $\Delta\rho$ is defined as SLD(ellipsoid) - SLD(solvent).  In the
parameters, $R_a$ is the minor equatorial radius, $R_b$ is the major
equatorial radius, and $R_c$ is the polar radius of the ellipsoid.

NB: The 2nd virial coefficient of the triaxial solid ellipsoid is calculated
after sorting the three radii to give the most appropriate prolate or oblate
form, from the new polar radius $R_p = R_c$ and effective equatorial radius,
$R_e = \sqrt{R_a R_b}$, to then be used as the effective radius for $S(q)$
when $P(q) \cdot S(q)$ is applied.

Validation
----------

Validation of our code was done by comparing the output of the
1D calculation to the angular average of the output of 2D calculation
over all possible angles.


References
----------

#. Finnigan, J.A., Jacobs, D.J., 1971. *Light scattering by ellipsoidal
   particles in solution*, J. Phys. D: Appl. Phys. 4, 72-77.
   doi:10.1088/0022-3727/4/1/310

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Last Modified by:** Paul Kienzle (improved calculation) **Date:** April 4, 2017
* **Last Reviewed by:** Paul Kienzle & Richard Heenan **Date:**  April 4, 2017
"""

import numpy as np
from numpy import inf, sin, cos, pi

name = "triaxial_ellipsoid"
title = "Ellipsoid of uniform scattering length density with three independent axes."

description = """
   Triaxial ellipsoid - see main documentation.
"""
category = "shape:ellipsoid"

#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["sld", "1e-6/Ang^2", 4, [-inf, inf], "sld",
               "Ellipsoid scattering length density"],
              ["sld_solvent", "1e-6/Ang^2", 1, [-inf, inf], "sld",
               "Solvent scattering length density"],
              ["radius_equat_minor", "Ang", 20, [0, inf], "volume",
               "Minor equatorial radius, Ra"],
              ["radius_equat_major", "Ang", 400, [0, inf], "volume",
               "Major equatorial radius, Rb"],
              ["radius_polar", "Ang", 10, [0, inf], "volume",
               "Polar radius, Rc"],
              ["theta", "degrees", 60, [-360, 360], "orientation",
               "polar axis to beam angle"],
              ["phi", "degrees", 60, [-360, 360], "orientation",
               "rotation about beam"],
              ["psi", "degrees", 60, [-360, 360], "orientation",
               "rotation about polar axis"],
             ]

source = ["lib/sas_3j1x_x.c", "lib/gauss76.c", "triaxial_ellipsoid.c"]
# Equations do not require Ra <= Rb <= Rc so don't test for it.
#valid = ("radius_equat_minor <= radius_equat_major"
#         " && radius_equat_major <= radius_polar")
have_Fq = True
radius_effective_modes = [
    "equivalent biaxial ellipsoid average curvature",
    "equivalent volume sphere", "min radius", "max radius",
    ]

def random():
    """Return a random parameter set for the model."""
    a, b, c = 10**np.random.uniform(1, 4.7, size=3)
    pars = dict(
        radius_equat_minor=a,
        radius_equat_major=b,
        radius_polar=c,
    )
    return pars

q = 0.1
# april 6 2017, rkh add unit tests
#     NOT compared with any other calc method, assume correct!
# check 2d test after pull #890
qx = q*cos(pi/6.0)
qy = q*sin(pi/6.0)
tests = [
    [{}, 0.05, 24.8839548033],
    [{'theta':80., 'phi':10.}, (qx, qy), 166.712060266],
    ]
del qx, qy  # not necessary to delete, but cleaner



#!/usr/bin/env python
"""
Asymmetric shape integration

Usage:

    explore/asymint.py [MODEL] [q-value]

Computes the numerical integral over theta and phi of the given model at a
single point q using different algorithms or the same algorithm with different
precision.  It also displays a 2-D image of the theta-phi surface that is
being integrated.

The available models are:

    triaxial_ellipsoid, parallelpiped, paracrystal, cylinder, sphere

Cylinder and sphere are included as simple checks on the integration
algorithms. Cylinder is better investigated using 1-D integration methods in
explore/symint.py.  Sphere has an easily computed analytic value which is
identical for all theta-phi for a given q, so it is useful for checking
that the normalization constants are correct for the different algorithms.
"""

from __future__ import print_function, division

import os, sys
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
import warnings

import numpy as np
import mpmath as mp
from numpy import pi, sin, cos, sqrt, exp, expm1, degrees, log10, arccos
from numpy.polynomial.legendre import leggauss
from scipy.integrate import dblquad, simps, romb, romberg
import pylab

import sasmodels.special as sp

DTYPE = 'd'

class MPenv:
    sqrt = staticmethod(mp.sqrt)
    exp = staticmethod(mp.exp)
    expm1 = staticmethod(mp.expm1)
    cos = staticmethod(mp.cos)
    sin = staticmethod(mp.sin)
    tan = staticmethod(mp.tan)
    @staticmethod
    def sas_3j1x_x(x):
        return 3*(mp.sin(x)/x - mp.cos(x))/(x*x)
    @staticmethod
    def sas_2J1x_x(x):
        return 2*mp.j1(x)/x
    @staticmethod
    def sas_sinx_x(x):
        return mp.sin(x)/x
    pi = mp.pi
    mpf = staticmethod(mp.mpf)

class NPenv:
    sqrt = staticmethod(np.sqrt)
    exp = staticmethod(np.exp)
    expm1 = staticmethod(np.expm1)
    cos = staticmethod(np.cos)
    sin = staticmethod(np.sin)
    tan = staticmethod(np.tan)
    sas_3j1x_x = staticmethod(sp.sas_3j1x_x)
    sas_2J1x_x = staticmethod(sp.sas_2J1x_x)
    sas_sinx_x = staticmethod(sp.sas_sinx_x)
    pi = np.pi
    #mpf = staticmethod(float)
    mpf = staticmethod(lambda x: np.array(x, DTYPE))

SLD = 3
SLD_SOLVENT = 6
CONTRAST = SLD - SLD_SOLVENT

# Carefully code models so that mpmath will use full precision.  That means:
#    * wrap inputs in env.mpf
#    * don't use floating point constants, only integers
#    * for division, make sure the numerator or denominator is env.mpf
#    * use env.pi, env.sas_sinx_x, etc. for functions
def make_parallelepiped(a, b, c, env=NPenv):
    a, b, c = env.mpf(a), env.mpf(b), env.mpf(c)
    def Fq(qa, qb, qc):
        siA = env.sas_sinx_x(a*qa/2)
        siB = env.sas_sinx_x(b*qb/2)
        siC = env.sas_sinx_x(c*qc/2)
        return siA * siB * siC
    Fq.__doc__ = "parallelepiped a=%g, b=%g c=%g"%(a, b, c)
    volume = a*b*c
    norm = CONTRAST**2*volume/10000
    return norm, Fq

def make_core_shell_parallelepiped(a, b, c, da, db, dc, slda, sldb, sldc, env=NPenv):
    overlapping = False
    a, b, c = env.mpf(a), env.mpf(b), env.mpf(c)
    da, db, dc = env.mpf(da), env.mpf(db), env.mpf(dc)
    slda, sldb, sldc = env.mpf(slda), env.mpf(sldb), env.mpf(sldc)
    dr0 = CONTRAST
    drA, drB, drC = slda-SLD_SOLVENT, sldb-SLD_SOLVENT, sldc-SLD_SOLVENT
    tA, tB, tC = a + 2*da, b + 2*db, c + 2*dc
    def Fq(qa, qb, qc):
        siA = a*env.sas_sinx_x(a*qa/2)
        siB = b*env.sas_sinx_x(b*qb/2)
        siC = c*env.sas_sinx_x(c*qc/2)
        siAt = tA*env.sas_sinx_x(tA*qa/2)
        siBt = tB*env.sas_sinx_x(tB*qb/2)
        siCt = tC*env.sas_sinx_x(tC*qc/2)
        if overlapping:
            return (dr0*siA*siB*siC
                    + drA*(siAt-siA)*siB*siC
                    + drB*siAt*(siBt-siB)*siC
                    + drC*siAt*siBt*(siCt-siC))
        else:
            return (dr0*siA*siB*siC
                    + drA*(siAt-siA)*siB*siC
                    + drB*siA*(siBt-siB)*siC
                    + drC*siA*siB*(siCt-siC))
    Fq.__doc__ = "core-shell parallelepiped a=%g, b=%g c=%g"%(a, b, c)
    if overlapping:
        volume = a*b*c + 2*da*b*c + 2*tA*db*c + 2*tA*tB*dc
    else:
        volume = a*b*c + 2*da*b*c + 2*a*db*c + 2*a*b*dc
    norm = 1/(volume*10000)
    return norm, Fq

def make_triaxial_ellipsoid(a, b, c, env=NPenv):
    a, b, c = env.mpf(a), env.mpf(b), env.mpf(c)
    def Fq(qa, qb, qc):
        qr = env.sqrt((a*qa)**2 + (b*qb)**2 + (c*qc)**2)
        return env.sas_3j1x_x(qr)
    Fq.__doc__ = "triaxial ellipsoid minor=%g, major=%g polar=%g"%(a, b, c)
    volume = 4*env.pi*a*b*c/3
    norm = CONTRAST**2*volume/10000
    return norm, Fq

def make_cylinder(radius, length, env=NPenv):
    radius, length = env.mpf(radius), env.mpf(length)
    def Fq(qa, qb, qc):
        qab = env.sqrt(qa**2 + qb**2)
        return env.sas_2J1x_x(qab*radius) * env.sas_sinx_x((qc*length)/2)
    Fq.__doc__ = "cylinder radius=%g, length=%g"%(radius, length)
    volume = env.pi*radius**2*length
    norm = CONTRAST**2*volume/10000
    return norm, Fq

def make_sphere(radius, env=NPenv):
    radius = env.mpf(radius)
    def Fq(qa, qb, qc):
        q = env.sqrt(qa**2 + qb**2 + qc**2)
        return env.sas_3j1x_x(q*radius)
    Fq.__doc__ = "sphere radius=%g"%(radius, )
    volume = 4*pi*radius**3
    norm = CONTRAST**2*volume/10000
    return norm, Fq

def make_paracrystal(radius, dnn, d_factor, lattice='bcc', env=NPenv):
    radius, dnn, d_factor = env.mpf(radius), env.mpf(dnn), env.mpf(d_factor)
    def sc(qa, qb, qc):
        return qa, qb, qc
    def bcc(qa, qb, qc):
        a1 = (+qa + qb + qc)/2
        a2 = (-qa - qb + qc)/2
        a3 = (-qa + qb - qc)/2
        return a1, a2, a3
    def fcc(qa, qb, qc):
        a1 = ( 0 + qb + qc)/2
        a2 = (-qa + 0 + qc)/2
        a3 = (-qa + qb + 0)/2
        return a1, a2, a3
    lattice_fn = {'sc': sc, 'bcc': bcc, 'fcc': fcc}[lattice]
    radius, dnn, d_factor = env.mpf(radius), env.mpf(dnn), env.mpf(d_factor)
    def Fq(qa, qb, qc):
        a1, a2, a3 = lattice_fn(qa, qb, qc)
        # Note: paper says that different directions can have different
        # distoration factors.  Easy enough to add to the code.
        arg = -(dnn*d_factor)**2*(a1**2 + a2**2 + a3**2)/2
        exp_arg = env.exp(arg)
        den = [((exp_arg - 2*env.cos(dnn*a))*exp_arg + 1) for a in (a1, a2, a3)]
        Sq = -env.expm1(2*arg)**3/(den[0]*den[1]*den[2])

        q = env.sqrt(qa**2 + qb**2 + qc**2)
        Fq = env.sas_3j1x_x(q*radius)
        # the caller computes F(q)**2, but we need it to compute S(q)*F(q)**2
        return env.sqrt(Sq)*Fq
    Fq.__doc__ = "%s paracrystal a=%g da=%g r=%g"%(lattice, dnn, d_factor, radius)
    def sphere_volume(r): return 4*env.pi*r**3/3
    Vf = {
        'sc': sphere_volume(radius/dnn),
        'bcc': 2*sphere_volume(env.sqrt(3)/2*radius/dnn),
        'fcc': 4*sphere_volume(1/env.sqrt(2)*radius/dnn),
    }[lattice]
    volume = sphere_volume(radius)
    norm = CONTRAST**2*volume/10000*Vf
    return norm, Fq

NORM = 1.0  # type: float
KERNEL = None  # type: CALLABLE[[ndarray, ndarray, ndarray], ndarray]
NORM_MP = 1  # type: mpf
KERNEL = None  # type: CALLABLE[[mpf, mpf, mpf], mpf]

SHAPES = [
    'sphere',
    'cylinder',
    'triaxial_ellipsoid',
    'parallelepiped',
    'core_shell_parallelepiped',
    'fcc_paracrystal',
    'bcc_paracrystal',
    'sc_paracrystal',
]
def build_shape(shape, **pars):
    global NORM, KERNEL
    global NORM_MP, KERNEL_MP

    # Note: using integer or string defaults for the sake of mpf
    if shape == 'sphere':
        RADIUS = pars.get('radius', 50)
        NORM, KERNEL = make_sphere(radius=RADIUS)
        NORM_MP, KERNEL_MP = make_sphere(radius=RADIUS, env=MPenv)
    elif shape == 'cylinder':
        #RADIUS, LENGTH = 10, 100000
        RADIUS = pars.get('radius', 10)
        LENGTH = pars.get('radius', 300)
        NORM, KERNEL = make_cylinder(radius=RADIUS, length=LENGTH)
        NORM_MP, KERNEL_MP = make_cylinder(radius=RADIUS, length=LENGTH, env=MPenv)
    elif shape == 'triaxial_ellipsoid':
        #A, B, C = 4450, 14000, 47
        A = pars.get('a', 445)
        B = pars.get('b', 140)
        C = pars.get('c', 47)
        NORM, KERNEL = make_triaxial_ellipsoid(A, B, C)
        NORM_MP, KERNEL_MP = make_triaxial_ellipsoid(A, B, C, env=MPenv)
    elif shape == 'parallelepiped':
        #A, B, C = 4450, 14000, 47
        A = pars.get('a', 445)
        B = pars.get('b', 140)
        C = pars.get('c', 47)
        NORM, KERNEL = make_parallelepiped(A, B, C)
        NORM_MP, KERNEL_MP = make_parallelepiped(A, B, C, env=MPenv)
    elif shape == 'core_shell_parallelepiped':
        #A, B, C = 4450, 14000, 47
        #A, B, C = 445, 140, 47  # integer for the sake of mpf
        A = pars.get('a', 114)
        B = pars.get('b', 1380)
        C = pars.get('c', 6800)
        DA = pars.get('da', 21)
        DB = pars.get('db', 58)
        DC = pars.get('dc', 2300)
        SLDA = pars.get('slda', "5")
        SLDB = pars.get('sldb', "-0.3")
        SLDC = pars.get('sldc', "11.5")
        ## default parameters from sasmodels
        #A,B,C,DA,DB,DC,SLDA,SLDB,SLDC = 400,75,35,10,10,10,2,4,2
        ## swap A-B-C to C-B-A
        #A, B, C, DA, DB, DC, SLDA, SLDB, SLDC = C, B, A, DC, DB, DA, SLDC, SLDB, SLDA
        #A,B,C,DA,DB,DC,SLDA,SLDB,SLDC = 10,20,30,100,200,300,1,2,3
        #SLD_SOLVENT,CONTRAST = 0, 4
        if 1: # C shortest
            B, C = C, B
            DB, DC = DC, DB
            SLDB, SLDC = SLDC, SLDB
        elif 0: # C longest
            A, C = C, A
            DA, DC = DC, DA
            SLDA, SLDC = SLDC, SLDA
        #NORM, KERNEL = make_core_shell_parallelepiped(A, B, C, DA, DB, DC, SLDA, SLDB, SLDC)
        NORM, KERNEL = make_core_shell_parallelepiped(A, B, C, DA, DB, DC, SLDA, SLDB, SLDC)
        NORM_MP, KERNEL_MP = make_core_shell_parallelepiped(A, B, C, DA, DB, DC, SLDA, SLDB, SLDC, env=MPenv)
    elif shape.endswith('paracrystal'):
        LATTICE, _ = shape.split('_')
        DNN = pars.get('dnn', 220)
        D_FACTOR = pars.get('d_factor', '0.06')
        RADIUS = pars.get('radius', 40)
        NORM, KERNEL = make_paracrystal(
            radius=RADIUS, dnn=DNN, d_factor=D_FACTOR, lattice=LATTICE)
        NORM_MP, KERNEL_MP = make_paracrystal(
            radius=RADIUS, dnn=DNN, d_factor=D_FACTOR, lattice=LATTICE, env=MPenv)
    else:
        raise ValueError("Unknown shape %r"%shape)

# Note: hardcoded in mp_quad
THETA_LOW, THETA_HIGH = 0, pi
PHI_LOW, PHI_HIGH = 0, 2*pi
SCALE = 1

# mathematica code for triaxial_ellipsoid (untested)
_ = """
R[theta_, phi_, a_, b_, c_] := Sqrt[(a Sin[theta]Cos[phi])^2 + (b Sin[theta]Sin[phi])^2 + (c Cos[theta])^2]
Sphere[q_, r_] := 3 SphericalBesselJ[q r]/(q r)
V[a_, b_, c_] := 4/3 pi a b c
Norm[sld_, solvent_, a_, b_, c_] := V[a, b, c] (solvent - sld)^2
F[q_, theta_, phi_, a_, b_, c_] := Sphere[q, R[theta, phi, a, b, c]]
I[q_, sld_, solvent_, a_, b_, c_] := Norm[sld, solvent, a, b, c]/(4 pi) Integrate[F[q, theta, phi, a, b, c]^2 Sin[theta], {phi, 0, 2 pi}, {theta, 0, pi}]
I[6/10^3, 63/10, 3, 445, 140, 47]
"""

# 2D integration functions
def mp_quad_2d(q):
    evals = [0]
    def integrand(theta, phi):
        evals[0] += 1
        qab = q*mp.sin(theta)
        qa = qab*mp.cos(phi)
        qb = qab*mp.sin(phi)
        qc = q*mp.cos(theta)
        Zq = KERNEL_MP(qa, qb, qc)**2
        return Zq*mp.sin(theta)
    ans = mp.quad(integrand, (0, mp.pi), (0, 2*mp.pi))
    Iq = NORM_MP*ans/(4*mp.pi)
    return evals[0], Iq

def kernel_2d(q, theta, phi):
    """
    S(q) kernel for paracrystal forms.
    """
    qab = q*sin(theta)
    qa = qab*cos(phi)
    qb = qab*sin(phi)
    qc = q*cos(theta)
    return NORM*KERNEL(qa, qb, qc)**2

def scipy_dblquad_2d(q):
    """
    Compute the integral using scipy dblquad.  This gets the correct answer
    eventually, but it is slow.
    """
    evals = [0]
    def integrand(phi, theta):
        evals[0] += 1
        Zq = kernel_2d(q, theta=theta, phi=phi)
        return Zq*sin(theta)
    ans = dblquad(integrand, THETA_LOW, THETA_HIGH, lambda x: PHI_LOW, lambda x: PHI_HIGH)[0]
    return evals[0], ans*SCALE/(4*pi)


def scipy_romberg_2d(q):
    """
    Compute the integral using romberg integration.  This function does not
    complete in a reasonable time.  No idea if it is accurate.
    """
    evals = [0]
    def inner(phi, theta):
        evals[0] += 1
        return kernel_2d(q, theta=theta, phi=phi)
    def outer(theta):
        Zq = romberg(inner, PHI_LOW, PHI_HIGH, divmax=100, args=(theta,))
        return Zq*sin(theta)
    ans = romberg(outer, THETA_LOW, THETA_HIGH, divmax=100)
    return evals[0], ans*SCALE/(4*pi)


def semi_romberg_2d(q, n=100):
    """
    Use 1D romberg integration in phi and regular simpsons rule in theta.
    """
    evals = [0]
    def inner(phi, theta):
        evals[0] += 1
        return kernel_2d(q, theta=theta, phi=phi)
    theta = np.linspace(THETA_LOW, THETA_HIGH, n)
    Zq = [romberg(inner, PHI_LOW, PHI_HIGH, divmax=100, args=(t,)) for t in theta]
    ans = simps(np.array(Zq)*sin(theta), dx=theta[1]-theta[0])
    return evals[0], ans*SCALE/(4*pi)

def gauss_quad_2d(q, n=150):
    """
    Compute the integral using gaussian quadrature for n = 20, 76 or 150.
    """
    z, w = leggauss(n)
    theta = (THETA_HIGH-THETA_LOW)*(z + 1)/2 + THETA_LOW
    phi = (PHI_HIGH-PHI_LOW)*(z + 1)/2 + PHI_LOW
    Atheta, Aphi = np.meshgrid(theta, phi)
    Aw = w[None, :] * w[:, None]
    sin_theta = abs(sin(Atheta))
    Zq = kernel_2d(q=q, theta=Atheta, phi=Aphi)
    # change from [-1,1] x [-1,1] range to [0, pi] x [0, 2 pi] range
    dxdy_stretch = (THETA_HIGH-THETA_LOW)/2 * (PHI_HIGH-PHI_LOW)/2
    Iq = np.sum(Zq*Aw*sin_theta)*SCALE/(4*pi) * dxdy_stretch
    return n**2, Iq

def gauss_quad_usub(q, n=150, dtype=DTYPE):
    """
    Compute the integral using gaussian quadrature for n = 20, 76 or 150.

    Use *u = sin theta* substitution, and restrict integration over a single
    quadrant for shapes that are mirror symmetric about AB, AC and BC planes.

    Note that this doesn't work for fcc/bcc paracrystals, which instead step
    over the entire 4 pi surface uniformly in theta-phi.
    """
    z, w = leggauss(n)
    cos_theta = 0.5 * (z + 1)
    theta = arccos(cos_theta)
    phi = pi/2*(0.5 * (z + 1))
    Atheta, Aphi = np.meshgrid(theta, phi)
    Aw = w[None, :] * w[:, None]
    q, Atheta, Aphi, Aw = [np.asarray(v, dtype=dtype) for v in (q, Atheta, Aphi, Aw)]
    Zq = kernel_2d(q=q, theta=Atheta, phi=Aphi)
    Iq = np.sum(Zq*Aw)*0.25
    return n**2, Iq

def gridded_2d(q, n=300):
    """
    Compute the integral on a regular grid using rectangular, trapezoidal,
    simpsons, and romberg integration.  Romberg integration requires that
    the grid be of size n = 2**k + 1.
    """
    theta = np.linspace(THETA_LOW, THETA_HIGH, n)
    phi = np.linspace(PHI_LOW, PHI_HIGH, n)
    Atheta, Aphi = np.meshgrid(theta, phi)
    Zq = kernel_2d(q=q, theta=Atheta, phi=Aphi)
    Zq *= abs(sin(Atheta))
    dx, dy = theta[1]-theta[0], phi[1]-phi[0]
    print("rect-%d"%n, n**2, np.sum(Zq)*dx*dy*SCALE/(4*pi))
    print("trapz-%d"%n, n**2, np.trapz(np.trapz(Zq, dx=dx), dx=dy)*SCALE/(4*pi))
    print("simpson-%d"%n, n**2, simps(simps(Zq, dx=dx), dx=dy)*SCALE/(4*pi))
    print("romb-%d"%n, n**2, romb(romb(Zq, dx=dx), dx=dy)*SCALE/(4*pi))

def quadpy_method(q, rule):
    """
    Use *rule*="name:index" where name and index are chosen from below.

    Available rule names and the corresponding indices::

        AlbrechtCollatz: [1-5]
        BazantOh: 9, 11, 13
        HeoXu: 13, 15, 17, 19-[1-2], 21-[1-6], 23-[1-3], 25-[1-2], 27-[1-3],
            29, 31, 33, 35, 37, 39-[1-2]
        FliegeMaier: 4, 9, 16, 25
        Lebedev: 3[a-c], 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 35,
            41, 47, 53, 59, 65, 71, 77 83, 89, 95, 101, 107, 113, 119, 125, 131
        McLaren: [1-10]
        Stroud: U3 3-1, U3 5-[1-5], U3 7-[1-2], U3 8-1, U3 9-[1-3],
            U3 11-[1-3], U3 14-1
    """
    try:
        import quadpy
    except ImportError:
        warnings.warn("use 'pip install quadpy' to enable quadpy.sphere tests")
        return

    from quadpy.sphere import (AlbrechtCollatz, BazantOh, HeoXu,
        FliegeMaier, Lebedev, McLaren, Stroud, integrate_spherical)
    RULES = {
        'AlbrechtCollatz': AlbrechtCollatz,
        'BazantOh': BazantOh,
        'HeoXu': HeoXu,
        'FliegeMaier': FliegeMaier,
        'Lebedev': Lebedev,
        'McLaren': McLaren,
        'Stroud': Stroud,
    }
    int_index = 'AlbrechtCollatz', 'McLaren'

    rule_name, rule_index = rule.split(':')
    index = int(rule_index) if rule_name in int_index else rule_index
    rule_obj = RULES[rule_name](index)
    fn = lambda azimuthal, polar: kernel_2d(q=q, theta=polar, phi=azimuthal)
    Iq = integrate_spherical(fn, rule=rule_obj)/(4*pi)
    print("%s degree=%d points=%s => %.15g"
          % (rule, rule_obj.degree, len(rule_obj.points), Iq))

def plot_2d(q, n=300):
    """
    Plot the 2D surface that needs to be integrated in order to compute
    the BCC S(q) at a particular q, dnn and d_factor.  *n* is the number
    of points in the grid.
    """
    theta = np.linspace(THETA_LOW, THETA_HIGH, n)
    phi = np.linspace(PHI_LOW, PHI_HIGH, n)
    Atheta, Aphi = np.meshgrid(theta, phi)
    Zq = kernel_2d(q=q, theta=Atheta, phi=Aphi)
    #Zq *= abs(sin(Atheta))
    pylab.pcolor(degrees(theta), degrees(phi), log10(np.fmax(Zq, 1.e-6)))
    pylab.axis('tight')
    pylab.title("%s I(q,t) sin(t) for q=%g" % (KERNEL.__doc__, q))
    pylab.xlabel("theta (degrees)")
    pylab.ylabel("phi (degrees)")
    cbar = pylab.colorbar()
    cbar.set_label('log10 S(q)')
    pylab.show()

def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="asymmetric integration explorer",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
    parser.add_argument('-s', '--shape', choices=SHAPES,
                        default='parallelepiped',
                        help='oriented shape')
    parser.add_argument('-q', '--q_value', type=str, default='0.005',
                        help='Q value to evaluate')
    parser.add_argument('pars', type=str, nargs='*', default=[],
                        help='p=val for p in shape parameters')
    opts = parser.parse_args()
    pars = {k: v for par in opts.pars for k, v in [par.split('=')]}
    build_shape(opts.shape, **pars)

    Q = float(opts.q_value)
    if opts.shape == 'sphere':
        print("exact", NORM*sp.sas_3j1x_x(Q*RADIUS)**2)

    # Methods from quadpy, if quadpy is available
    #  AlbrechtCollatz: [1-5]
    #  BazantOh: 9, 11, 13
    #  HeoXu: 13, 15, 17, 19-[1-2], 21-[1-6], 23-[1-3], 25-[1-2], 27-[1-3],
    #     29, 31, 33, 35, 37, 39-[1-2]
    #  FliegeMaier: 4, 9, 16, 25
    #  Lebedev: 3[a-c], 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 35,
    #     41, 47, 53, 59, 65, 71, 77 83, 89, 95, 101, 107, 113, 119, 125, 131
    #  McLaren: [1-10]
    #  Stroud: U3 3-1, U3 5-[1-5], U3 7-[1-2], U3 8-1, U3 9-[1-3],
    #     U3 11-[1-3], U3 14-1
    quadpy_method(Q, "AlbrechtCollatz:5")
    quadpy_method(Q, "HeoXu:39-2")
    quadpy_method(Q, "FliegeMaier:25")
    quadpy_method(Q, "Lebedev:19")
    quadpy_method(Q, "Lebedev:131")
    quadpy_method(Q, "McLaren:10")
    quadpy_method(Q, "Stroud:U3 14-1")

    print("gauss-20 points=%d => %.15g" % gauss_quad_2d(Q, n=20))
    print("gauss-76 points=%d => %.15g" % gauss_quad_2d(Q, n=76))
    print("gauss-150 points=%d => %.15g" % gauss_quad_2d(Q, n=150))
    print("gauss-500 points=%d => %.15g" % gauss_quad_2d(Q, n=500))
    print("gauss-1025 points=%d => %.15g" % gauss_quad_2d(Q, n=1025))
    print("gauss-2049 points=%d => %.15g" % gauss_quad_2d(Q, n=2049))
    print("gauss-20 usub points=%d => %.15g" % gauss_quad_usub(Q, n=20))
    print("gauss-76 usub points=%d => %.15g" % gauss_quad_usub(Q, n=76))
    print("gauss-150 usub points=%d => %.15g" % gauss_quad_usub(Q, n=150))

    #gridded_2d(Q, n=2**8+1)
    gridded_2d(Q, n=2**10+1)
    #gridded_2d(Q, n=2**12+1)
    #gridded_2d(Q, n=2**15+1)
    # adaptive forms on models for which the calculations are fast enough
    SLOW_SHAPES = {
        'fcc_paracrystal', 'bcc_paracrystal', 'sc_paracrystal',
        'core_shell_parallelepiped',
    }
    if opts.shape not in SLOW_SHAPES:
        print("dblquad", *scipy_dblquad_2d(Q))
        print("semi-romberg-100", *semi_romberg_2d(Q, n=100))
        print("romberg", *scipy_romberg_2d(Q))
        with mp.workprec(100):
            print("mpmath", *mp_quad_2d(mp.mpf(opts.q_value)))
    plot_2d(Q, n=200)

if __name__ == "__main__":
    main()



"""
Custom Models
-------------

This is a place holder for the custom models namespace.  When models are
loaded from a file by :func:`sasmodels.generate.load_kernel_module` they are
loaded as if they exist in *sasmodels.custom*.  This package needs to exist
for this to occur without error.
"""
from __future__ import division, print_function

import sys
import os
from os.path import basename, splitext, join as joinpath, exists, dirname

# pylint: disable=unused-import
try:
    from typing import Dict, List, Tuple
    from types import ModuleType
except ImportError:
    pass
# pylint: enable=unused-import


try:
    # Python 3.5 and up
    from importlib.util import spec_from_file_location, module_from_spec  # type: ignore
    def load_module_from_path(fullname, path):
        # type: (str, str) -> ModuleType
        """load module from *path* as *fullname*"""
        spec = spec_from_file_location(fullname, os.path.expanduser(path))
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        return module
except ImportError:
    # CRUFT: python 2
    import imp
    def load_module_from_path(fullname, path):
        # type: (str, str) -> ModuleType
        """load module from *path* as *fullname*"""
        # Clear out old definitions, if any
        if fullname in sys.modules:
            del sys.modules[fullname]
        if path.endswith(".py") and os.path.exists(path) and os.path.exists(path+"c"):
            # remove automatic pyc file before loading a py file
            os.unlink(path+"c")
        module = imp.load_source(fullname, os.path.expanduser(path))
        return module

_MODULE_CACHE = {}  # type: Dict[str, Tuple[ModuleType, int]]
_MODULE_DEPENDS = {}  # type: Dict[str, List[str]]
_MODULE_DEPENDS_STACK = []  # type: List[str]
def load_custom_kernel_module(path):
    # type: (str) -> ModuleType
    """load SAS kernel from *path* as *sasmodels.custom.modelname*"""
    # Pull off the last .ext if it exists; there may be others
    name = basename(splitext(path)[0])
    path = os.path.expanduser(path)

    # Reload module if necessary.
    if need_reload(path):
        # Assume the module file is the only dependency
        _MODULE_DEPENDS[path] = set([path])

        # Load the module while pushing it onto the dependency stack.  If
        # this triggers any submodules, then they will add their dependencies
        # to this module as the "working_on" parent.  Pop the stack when the
        # module is loaded.
        _MODULE_DEPENDS_STACK.append(path)
        module = load_module_from_path('sasmodels.custom.'+name, path)
        _MODULE_DEPENDS_STACK.pop()

        # Include external C code in the dependencies.  We are looking
        # for module.source and assuming that it is a list of C source files
        # relative to the module itself.  Any files that do not exist,
        # such as those in the standard libraries, will be ignored.
        # TODO: look in builtin module path for standard c sources
        # TODO: share code with generate.model_sources
        c_sources = getattr(module, 'source', None)
        if isinstance(c_sources, (list, tuple)):
            _MODULE_DEPENDS[path].update(_find_sources(path, c_sources))

        # Cache the module, and tag it with the newest timestamp
        timestamp = max(os.path.getmtime(f) for f in _MODULE_DEPENDS[path])
        _MODULE_CACHE[path] = module, timestamp

        #print("loading", os.path.basename(path), _MODULE_CACHE[path][1],
        #    [os.path.basename(p) for p in _MODULE_DEPENDS[path]])

    # Add path and all its dependence to the parent module, if there is one.
    if _MODULE_DEPENDS_STACK:
        working_on = _MODULE_DEPENDS_STACK[-1]
        _MODULE_DEPENDS[working_on].update(_MODULE_DEPENDS[path])

    return _MODULE_CACHE[path][0]

def need_reload(path):
    # type: (str) -> bool
    """
    Return True if any path dependencies have a timestamp newer than the time
    when the path was most recently loaded.
    """
    # TODO: fails if a dependency has a modification time in the future
    # If the newest dependency has a time stamp in the future, then this
    # will be recorded as the cached time.  When a second dependency
    # is updated to the current time stamp, it will still be considered
    # older than the current build and the reload will not be triggered.
    # Could instead treat all future times as 0 here and in the code above
    # which records the newest timestamp.  This will force a reload when
    # the future time is reached, but other than that should perform
    # correctly.  Probably not worth the extra code...
    _, cache_time = _MODULE_CACHE.get(path, (None, -1))
    depends = _MODULE_DEPENDS.get(path, [path])
    #print("reload", any(cache_time < os.path.getmtime(p) for p in depends))
    #for f in depends: print(">>>  ", f, os.path.getmtime(f))
    return any(cache_time < os.path.getmtime(p) for p in depends)

def _find_sources(path, source_list):
    # type: (str, List[str]) -> List[str]
    """
    Return a list of the sources relative to base file; ignore any that
    are not found.
    """
    root = dirname(path)
    found = []
    for source_name in source_list:
        source_path = joinpath(root, source_name)
        if exists(source_path):
            found.append(source_path)
    return found



"""
Kernel Call Details
===================

When calling sas computational kernels with polydispersity there are a
number of details that need to be sent to the caller.  This includes the
list of polydisperse parameters, the number of points in the polydispersity
weight distribution, and which parameter is the "theta" parameter for
polar coordinate integration.  The :class:`CallDetails` object maintains
this data.  Use :func:`make_details` to build a *details* object which
can be passed to one of the computational kernels.
"""

from __future__ import print_function

import numpy as np  # type: ignore
from numpy import cos, sin, radians

from .modelinfo import NUM_MAGNETIC_PARS, NUM_COMMON_PARS

try:
    np.meshgrid([])
    meshgrid = np.meshgrid
except Exception:
    # CRUFT: np.meshgrid requires multiple vectors
    def meshgrid(*args):
        """See docs from a recent version of numpy"""
        if len(args) > 1:
            return np.meshgrid(*args)
        else:
            return [np.asarray(v) for v in args]

# pylint: disable=unused-import
try:
    from typing import List, Tuple, Sequence
    from .modelinfo import ModelInfo, ParameterTable
    from .kernel import Kernel
except ImportError:
    pass
# pylint: enable=unused-import


class CallDetails(object):
    """
    Manage the polydispersity information for the kernel call.

    Conceptually, a polydispersity calculation is an integral over a mesh
    in n-D space where n is the number of polydisperse parameters.  In order
    to keep the program responsive, and not crash the GPU, only a portion
    of the mesh is computed at a time.  Meshes with a large number of points
    will therefore require many calls to the polydispersity loop.  Restarting
    a nested loop in the middle requires that the indices of the individual
    mesh dimensions can be computed for the current loop location.  This
    is handled by the *pd_stride* vector, with n//stride giving the loop
    index and n%stride giving the position in the sub loops.

    One of the parameters may be the latitude.  When integrating in polar
    coordinates, the total circumference decreases as latitude varies from
    pi r^2 at the equator to 0 at the pole, and the weight associated
    with a range of latitude values needs to be scaled by this circumference.
    This scale factor needs to be updated each time the theta value
    changes.  *theta_par* indicates which of the values in the parameter
    vector is the latitude parameter, or -1 if there is no latitude
    parameter in the model.  In practice, the normalization term cancels
    if the latitude is not a polydisperse parameter.
    """
    parts = None  # type: List["CallDetails"]
    def __init__(self, model_info):
        # type: (ModelInfo) -> None
        parameters = model_info.parameters
        max_pd = parameters.max_pd

        # Structure of the call details buffer:
        #   pd_par[max_pd]     pd params in order of length
        #   pd_length[max_pd]  length of each pd param
        #   pd_offset[max_pd]  offset of pd values in parameter array
        #   pd_stride[max_pd]  index of pd value in loop = n//stride[k]
        #   num_eval           total length of pd loop
        #   num_weights        total length of the weight vector
        #   num_active         number of pd params
        #   theta_par          parameter number for theta parameter
        self.buffer = np.empty(4*max_pd + 4, 'i4')

        # generate views on different parts of the array
        self._pd_par = self.buffer[0 * max_pd:1 * max_pd]
        self._pd_length = self.buffer[1 * max_pd:2 * max_pd]
        self._pd_offset = self.buffer[2 * max_pd:3 * max_pd]
        self._pd_stride = self.buffer[3 * max_pd:4 * max_pd]

        # theta_par is fixed
        self.theta_par = parameters.theta_offset

        # offset and length are for all parameters, not just pd parameters
        # They are not sent to the kernel function, though they could be.
        # They are used by the composite models (sum and product) to
        # figure out offsets into the combined value list.
        self.offset = None  # type: np.ndarray
        self.length = None  # type: np.ndarray

        # keep hold of ifno show() so we can break a values vector
        # into the individual components
        self.info = model_info

    @property
    def pd_par(self):
        """List of polydisperse parameters"""
        return self._pd_par

    @property
    def pd_length(self):
        """Number of weights for each polydisperse parameter"""
        return self._pd_length

    @property
    def pd_offset(self):
        """Offsets for the individual weight vectors in the set of weights"""
        return self._pd_offset

    @property
    def pd_stride(self):
        """Stride in the pd mesh for each pd dimension"""
        return self._pd_stride

    @property
    def num_eval(self):
        """Total size of the pd mesh"""
        return self.buffer[-4]

    @num_eval.setter
    def num_eval(self, v):
        """Total size of the pd mesh"""
        self.buffer[-4] = v

    @property
    def num_weights(self):
        """Total length of all the weight vectors"""
        return self.buffer[-3]

    @num_weights.setter
    def num_weights(self, v):
        """Total length of all the weight vectors"""
        self.buffer[-3] = v

    @property
    def num_active(self):
        """Number of active polydispersity loops"""
        return self.buffer[-2]

    @num_active.setter
    def num_active(self, v):
        """Number of active polydispersity loops"""
        self.buffer[-2] = v

    @property
    def theta_par(self):
        """Location of the theta parameter in the parameter vector"""
        return self.buffer[-1]

    @theta_par.setter
    def theta_par(self, v):
        """Location of the theta parameter in the parameter vector"""
        self.buffer[-1] = v

    def show(self, values=None):
        """Print the polydispersity call details to the console"""
        print("===== %s details ===="%self.info.name)
        print("num_active:%d  num_eval:%d  num_weights:%d  theta=%d"
              % (self.num_active, self.num_eval, self.num_weights, self.theta_par))
        if self.pd_par.size:
            print("pd_par", self.pd_par)
            print("pd_length", self.pd_length)
            print("pd_offset", self.pd_offset)
            print("pd_stride", self.pd_stride)
        if values is not None:
            nvalues = self.info.parameters.nvalues
            print("scale, background", values[:2])
            print("val", values[2:nvalues])
            print("pd", values[nvalues:nvalues+self.num_weights])
            print("wt", values[nvalues+self.num_weights:nvalues+2*self.num_weights])
            print("offsets", self.offset)


def make_details(model_info, length, offset, num_weights):
    # type: (ModelInfo, np.ndarray, np.ndarray, int) -> CallDetails
    """
    Return a :class:`CallDetails` object for a polydisperse calculation
    of the model defined by *model_info*.  Polydispersity is defined by
    the *length* of the polydispersity distribution for each parameter
    and the *offset* of the distribution in the polydispersity array.
    Monodisperse parameters should use a polydispersity length of one
    with weight 1.0. *num_weights* is the total length of the polydispersity
    array.
    """
    #pars = model_info.parameters.call_parameters[2:model_info.parameters.npars+2]
    #print(", ".join(str(i)+"-"+p.id for i,p in enumerate(pars)))
    #print("len:",length)
    #print("off:",offset)

    # Check that we aren't using too many polydispersity loops
    num_active = np.sum(length > 1)
    max_pd = model_info.parameters.max_pd
    if num_active > max_pd:
        raise ValueError("Too many polydisperse parameters")

    # Decreasing list of polydpersity lengths
    # Note: the reversing view, x[::-1], does not require a copy
    idx = np.argsort(length)[::-1][:max_pd]
    pd_stride = np.cumprod(np.hstack((1, length[idx])))

    call_details = CallDetails(model_info)
    call_details.pd_par[:max_pd] = idx
    call_details.pd_length[:max_pd] = length[idx]
    call_details.pd_offset[:max_pd] = offset[idx]
    call_details.pd_stride[:max_pd] = pd_stride[:-1]
    call_details.num_eval = pd_stride[-1]
    call_details.num_weights = num_weights
    call_details.num_active = num_active
    call_details.length = length
    call_details.offset = offset
    #call_details.show()
    return call_details


ZEROS = tuple([0.]*31)
def make_kernel_args(kernel, mesh):
    # type: (Kernel, Tuple[List[np.ndarray], List[np.ndarray]]) -> Tuple[CallDetails, np.ndarray, bool]
    """
    Converts (value, dispersity, weight) for each parameter into kernel pars.

    Returns a CallDetails object indicating the polydispersity, a data object
    containing the different values, and the magnetic flag indicating whether
    any magnetic magnitudes are non-zero. Magnetic vectors (M0, phi, theta) are
    converted to rectangular coordinates (mx, my, mz).
    """
    npars = kernel.info.parameters.npars
    nvalues = kernel.info.parameters.nvalues
    scalars = [value for value, dispersity, weight in mesh]
    # skipping scale and background when building values and weights
    _, dispersity, weight = (
        zip(*mesh[NUM_COMMON_PARS:npars+NUM_COMMON_PARS]) if npars
        else ((), (), ()))
    #weight = correct_theta_weights(kernel.info.parameters, dispersity, weight)
    length = np.array([len(w) for w in weight])
    offset = np.cumsum(np.hstack((0, length)))
    call_details = make_details(kernel.info, length, offset[:-1], offset[-1])
    # Pad value array to a 32 value boundary
    data_len = nvalues + 2*sum(len(v) for v in dispersity)
    extra = (32 - data_len%32)%32
    data = np.hstack((scalars,) + dispersity + weight + ZEROS[:extra])
    data = data.astype(kernel.dtype)
    is_magnetic = convert_magnetism(kernel.info.parameters, data)
    #call_details.show()
    #print("data", data)
    return call_details, data, is_magnetic

def correct_theta_weights(parameters, dispersity, weights):
    # type: (ParameterTable, Sequence[np.ndarray], Sequence[np.ndarray]) -> Sequence[np.ndarray]
    """
    **Deprecated** Theta weights will be computed in the kernel wrapper if
    they are needed.

    If there is a theta parameter, update the weights of that parameter so that
    the cosine weighting required for polar integration is preserved.

    Avoid evaluation strictly at the pole, which would otherwise send the
    weight to zero.  This is probably not a problem in practice (if dispersity
    is +/- 90, then you probably should be using a 1-D model of the circular
    average).

    Note: scale and background parameters are not include in the tuples for
    dispersity and weights, so index is parameters.theta_offset, not
    parameters.theta_offset+2

    Returns updated weights vectors
    """
    # Apparently the parameters.theta_offset similarly skips scale and
    # and background, so the indexing works out, but they are still shipped
    # to the kernel, so we need to add two there.
    if parameters.theta_offset >= 0:
        index = parameters.theta_offset
        theta = dispersity[index]
        theta_weight = abs(cos(radians(theta)))
        weights = tuple(theta_weight*w if k == index else w
                        for k, w in enumerate(weights))
    return weights


def convert_magnetism(parameters, values):
    # type: (ParameterTable, Sequence[np.ndarray]) -> bool
    """
    Convert magnetism values from polar to rectangular coordinates.

    Returns True if any magnetism is present.
    """
    nmagpars = NUM_MAGNETIC_PARS*parameters.nmagnetic
    mag = values[parameters.nvalues-nmagpars : parameters.nvalues]
    mag = mag.reshape(-1, NUM_MAGNETIC_PARS)
    if np.any(mag[:, 0] != 0.0):
        M0 = mag[:, 0].copy()
        theta, phi = radians(mag[:, 1]), radians(mag[:, 2])
        mag[:, 0] = M0 * sin(theta) * cos(phi)  # mx
        mag[:, 1] = M0 * sin(theta) * sin(phi)  # my
        mag[:, 2] = M0 * cos(theta) # mz
        return True
    else:
        return False


def dispersion_mesh(model_info, mesh):
    # type: (ModelInfo, List[Tuple[float, np.ndarray, np.ndarray]]) -> Tuple[List[np.ndarray], List[np.ndarray]]
    """
    Create a mesh grid of dispersion parameters and weights.

    *mesh* is a list of (value, dispersity, weights), where the values
    are the individual parameter values, and (dispersity, weights) is
    the distribution of parameter values.

    Only the volume parameters should be included in this list.  Orientation
    parameters do not affect the calculation of effective radius or volume
    ratio.  This is convenient since it avoids the distinction between
    value and dispersity that is present in orientation parameters but not
    shape parameters.

    Returns [p1,p2,...],w where pj is a vector of values for parameter j
    and w is a vector containing the products for weights for each
    parameter set in the vector.
    """
    _, dispersity, weight = zip(*mesh)
    #weight = [w if len(w)>0 else [1.] for w in weight]
    weight = np.vstack([v.flatten() for v in meshgrid(*weight)])
    weight = np.prod(weight, axis=0)
    dispersity = [v.flatten() for v in meshgrid(*dispersity)]
    lengths = [par.length for par in model_info.parameters.kernel_parameters
               if par.type == 'volume']
    if any(n > 1 for n in lengths):
        pars = []
        offset = 0
        for n in lengths:
            pars.append(np.vstack(dispersity[offset:offset+n])
                        if n > 1 else dispersity[offset])
            offset += n
        dispersity = pars
    return dispersity, weight



"""
Parameter conversion table

*CONVERSION_TABLE* gives the old model name and a dictionary of old parameter
names for each parameter in sasmodels.  This is used by :mod:`.convert` to
determine the equivalent parameter set when comparing a sasmodels model to
the models defined in previous versions of SasView and sasmodels. This is now
versioned based on the version number of SasView.

When any sasmodels parameter or model name is changed, this must be modified to
account for that.

Usage::

    <old_Sasview_version> : {
        <new_model_name> : [
            <old_model_name> ,
            {
                <new_param_name_1> : <old_param_name_1>,
                ...
                <new_param_name_n> : <old_param_name_n>
            }
        ]
    }

Any future parameter and model name changes can and should be given in this
table for future compatibility.
"""

CONVERSION_TABLE = {
    (3, 1, 2) : {
        "adsorbed_layer": [
            "Core2ndMomentModel",
            {
                "scale": "scale",
                "second_moment": "second_moment",
                "density_shell": "density_poly",
                "sld_solvent": "sld_solv",
                "radius": "radius_core",
                "volfraction": "volf_cores",
                "background": "background",
                "adsorbed_amount": "ads_amount",
                "sld_shell": "sld_poly"
            }
        ],
        "barbell": [
            "BarBellModel",
            {
                "sld": "sld_barbell",
                "length": "len_bar",
                "radius_bell": "rad_bell",
                "radius": "rad_bar",
                "sld_solvent": "sld_solv"
            }
        ],
        "bcc_paracrystal": [
            "BCCrystalModel",
            {
                "sld": "sldSph",
                "sld_solvent": "sldSolv"
            }
        ],
        "be_polyelectrolyte": [
            "BEPolyelectrolyte",
            {
                "ionization_degree": "alpha",
                "polymer_concentration": "c",
                "salt_concentration": "cs",
                "virial_param": "h",
                "background": "background",
                "contrast_factor": "k",
                "bjerrum_length": "lb",
                "monomer_length": "b"
            }
        ],
        "binary_hard_sphere": [
            "BinaryHSModel",
            {
                "sld_sm": "ss_sld",
                "sld_lg": "ls_sld",
                "volfraction_sm": "vol_frac_ss",
                "radius_lg": "l_radius",
                "radius_sm": "s_radius",
                "volfraction_lg": "vol_frac_ls",
                "sld_solvent": "solvent_sld"
            }
        ],
        "broad_peak": [
            "BroadPeakModel",
            {
                "peak_pos": "q_peak",
                "scale": None,
                "lorentz_length": "length_l",
                "porod_scale": "scale_p",
                "lorentz_exp": "exponent_l",
                "lorentz_scale": "scale_l",
                "porod_exp": "exponent_p"
            }
        ],
        "capped_cylinder": [
            "CappedCylinderModel",
            {
                "sld": "sld_capcyl",
                "length": "len_cyl",
                "radius_cap": "rad_cap",
                "radius": "rad_cyl",
                "sld_solvent": "sld_solv"
            }
        ],
        "core_multi_shell": [
            "CoreMultiShellModel",
            {
                "thickness": "thick_shell",
                "sld": "sld_shell",
                "radius": "rad_core0",
                "sld_core": "sld_core0",
                "sld_solvent": "sld_solv",
                "n": "n_shells",
                "M0:sld_core": "M0_sld_core0",
                "mtheta:sld_core": "M_theta_core0",
                "mphi:sld_core": "M_phi_core0",
                "M0:sld1": "M0_sld_shell1",
                "mtheta:sld1": "M_theta_shell1",
                "mphi:sld1": "M_phi_shell1",
                "M0:sld2": "M0_sld_shell2",
                "mtheta:sld2": "M_theta_shell2",
                "mphi:sld2": "M_phi_shell2",
                "M0:sld3": "M0_sld_shell3",
                "mtheta:sld3": "M_theta_shell3",
                "mphi:sld3": "M_phi_shell3",
                "M0:sld4": "M0_sld_shell4",
                "mtheta:sld4": "M_theta_shell4",
                "mphi:sld4": "M_phi_shell4",
                "M0:sld_solvent": "M0_sld_solv",
                "mtheta:sld_solvent": "M_theta_solv",
                "mphi:sld_solvent": "M_phi_solv",
                "up:frac_i": "up_frac_i",
                "up:frac_f": "up_frac_f",
                "up:angle": "up_angle",
            }
        ],
        "core_shell_bicelle": [
            "CoreShellBicelleModel",
            {
                "phi": "axis_phi",
                "sld_core": "core_sld",
                "sld_rim": "rim_sld",
                "thick_face": "face_thick",
                "sld_solvent": "solvent_sld",
                "thick_rim": "rim_thick",
                "sld_face": "face_sld",
                "theta": "axis_theta"
            }
        ],
        "core_shell_cylinder": [
            "CoreShellCylinderModel",
            {
                "theta": "axis_theta",
                "phi": "axis_phi",
                "sld_shell": "shell_sld",
                "sld_solvent": "solvent_sld",
                "sld_core": "core_sld"
            }
        ],
        "core_shell_ellipsoid:1": [
            "CoreShellEllipsoidModel",
            {
                "sld_core": "sld_core",
                "sld_shell": "sld_shell",
                "sld_solvent": "sld_solvent",
                "radius_equat_core": "equat_core",
                "x_core": "polar_core",
                "thick_shell": "equat_shell",
                "x_polar_shell": "polar_shell",
                "theta": "axis_theta",
                "phi": "axis_phi",
            }
        ],
        "core_shell_ellipsoid": [
            "CoreShellEllipsoidXTModel",
            {
                "sld_core": "sld_core",
                "sld_shell": "sld_shell",
                "sld_solvent": "sld_solvent",
                "radius_equat_core": "equat_core",
                "thick_shell": "T_shell",
                "x_core": "X_core",
                "x_polar_shell": "XpolarShell",
                "theta": "axis_theta",
                "phi": "axis_phi",
            }
        ],
        "core_shell_parallelepiped": [
            "CSParallelepipedModel",
            {
                "sld_core": "sld_pcore",
                "sld_a": "sld_rimA",
                "sld_b": "sld_rimB",
                "sld_c": "sld_rimC",
                "sld_solvent": "sld_solv",
                "length_a": "shortA",
                "length_b": "midB",
                "length_c": "longC",
                "thick_rim_a": "rimA",
                "thick_rim_c": "rimC",
                "thick_rim_b": "rimB",
                "theta": "parallel_theta",
                "phi": "parallel_phi",
                "psi": "parallel_psi",
            }
        ],
        "core_shell_sphere": [
            "CoreShellModel",
            {
                "sld_core": "core_sld",
                "sld_shell": "shell_sld",
                "sld_solvent": "solvent_sld",
                "M0:sld_core": "M0_sld_core",
                "mtheta:sld_core": "M_theta_core",
                "mphi:sld_core": "M_phi_core",
                "M0:sld_shell": "M0_sld_shell",
                "mtheta:sld_shell": "M_theta_shell",
                "mphi:sld_shell": "M_phi_shell",
                "M0:sld_solvent": "M0_sld_solv",
                "mtheta:sld_solvent": "M_theta_solv",
                "mphi:sld_solvent": "M_phi_solv",
                "up:frac_i": "up_frac_i",
                "up:frac_f": "up_frac_f",
                "up:angle": "up_angle"
            }
        ],
        "correlation_length": [
            "CorrLength",
            {
                "porod_scale": "scale_p",
                "lorentz_scale": "scale_l",
                "porod_exp": "exponent_p",
                "lorentz_exp": "exponent_l",
                "cor_length": "length_l"
            },
            "CorrLengthModel"
        ],
        "cylinder": [
            "CylinderModel",
            {
                "sld": "sldCyl",
                "theta": "cyl_theta",
                "phi": "cyl_phi",
                "sld_solvent": "sldSolv",
                "M0:sld": "M0_sld_cyl",
                "mtheta:sld": "M_theta_cyl",
                "mphi:sld": "M_phi_cyl",
                "M0:sld_solvent": "M0_sld_solv",
                "mtheta:sld_solvent": "M_theta_solv",
                "mphi:sld_solvent": "M_phi_solv",
                "up:frac_i": "up_frac_i",
                "up:frac_f": "up_frac_f",
                "up:angle": "up_angle"
            }
        ],
        "dab": [
            "DABModel",
            {
                "cor_length": "length"
            }
        ],
        "ellipsoid": [
            "EllipsoidModel",
            {
                "phi": "axis_phi",
                "radius_equatorial": "radius_b",
                "sld": "sldEll",
                "theta": "axis_theta",
                "radius_polar": "radius_a",
                "sld_solvent": "sldSolv"
            }
        ],
        "elliptical_cylinder": [
            "EllipticalCylinderModel",
            {
                "axis_ratio": "r_ratio",
                "radius_minor": "r_minor",
                "sld": "sldCyl",
                "sld_solvent": "sldSolv",
                "theta": "cyl_theta",
                "phi": "cyl_phi",
                "psi": "cyl_psi",
            }
        ],
        "fcc_paracrystal": [
            "FCCrystalModel",
            {
                "sld": "sldSph",
                "sld_solvent": "sldSolv"
            }
        ],
        "flexible_cylinder": [
            "FlexibleCylinderModel",
            {
                "sld": "sldCyl",
                "sld_solvent": "sldSolv"
            }
        ],
        "flexible_cylinder_elliptical": [
            "FlexCylEllipXModel",
            {
                "sld": "sldCyl",
                "sld_solvent": "sldSolv"
            }
        ],
        "fractal": [
            "FractalModel",
            {
                "sld_block": "sldBlock",
                "radius": "radius",
                "cor_length": "cor_length",
                "sld_solvent": "sldSolv",
                "fractal_dim": "fractal_dim"
            }
        ],
        "fractal_core_shell": [
            "FractalCoreShell",
            {
                "sld_core": "core_sld",
                "sld_shell": "shell_sld",
                "sld_solvent": "solvent_sld",
                "radius": "radius",
                "thickness": "thickness",
                "fractal_dim": "frac_dim",
                "cor_length": "cor_length",
                "volfraction": "volfraction",
            },
            "FractalCoreShellModel"
        ],
        "fuzzy_sphere": [
            "FuzzySphereModel",
            {
                "sld": "sldSph",
                "fuzziness": "fuzziness",
                "radius": "radius",
                "sld_solvent": "sldSolv"
            }
        ],
        "gauss_lorentz_gel": [
            "GaussLorentzGel",
            {
                "gauss_scale": "scale_g",
                "cor_length_dynamic": "dyn_colength",
                "cor_length_static": "stat_colength",
                "background": "background",
                "lorentz_scale": "scale_l"
            },
            "GaussLorentzGelModel"
        ],
        "gaussian_peak": [
            "Peak Gauss Model",
            {
                "peak_pos": "q0",
                "sigma": "B",
            },
            "PeakGaussModel",
        ],
        "gel_fit": [
            "GelFitModel",
            {
                "rg": "radius",
                "lorentz_scale": "lScale",
                "guinier_scale": "gScale",
                "fractal_dim": "FractalExp",
                "cor_length": "zeta",
            }
        ],
        "guinier": [
            "Guinier",
            {
                "rg": "rg"
            },
            "GuinierModel",
        ],
        "guinier_porod": [
            "GuinierPorod",
            {
                "s": "dim",
                "rg": "rg",
                "porod_exp": "m",
                "scale": "scale",
                "background": "background"
            },
            "GuinierPorodModel",
        ],
        "hardsphere": [
            "HardsphereStructure",
            {
                "scale": "scale_factor",
                "radius_effective": "effect_radius",
            }
        ],
        "hayter_msa": [
            "HayterMSAStructure",
            {
                "scale": "scale_factor",
                "radius_effective": "effect_radius",
                "volfraction": "volfraction",
                "charge": "charge",
                "temperature": "temperature",
                "concentration_salt": "saltconc",
                "dielectconst": "dielectconst",
            }
        ],
        "hollow_cylinder": [
            "HollowCylinderModel",
            {
                "sld": "sldCyl",
                "sld_solvent": "sldSolv",
                "radius": "core_radius",
                "thickness": "radius",
                "length": "length",
                "theta": "axis_theta",
                "phi": "axis_phi",
            }
        ],
        "hollow_rectangular_prism": [
            "RectangularHollowPrismModel",
            {
                "sld": "sldPipe",
                "sld_solvent": "sldSolv",
                "length_a": "short_side",
                "b2a_ratio": "b2a_ratio",
                "c2a_ratio": "c2a_ratio",
                "thickness": "thickness",
            }
        ],
        "hollow_rectangular_prism_thin_walls": [
            "RectangularHollowPrismInfThinWallsModel",
            {
                "sld": "sldPipe",
                "sld_solvent": "sldSolv",
                "length_a": "short_side",
                "b2a_ratio": "b2a_ratio",
                "c2a_ratio": "c2a_ratio",
            }
        ],
        "lamellar": [
            "LamellarModel",
            {
                "sld": "sld_bi",
                "sld_solvent": "sld_sol",
                "thickness": "bi_thick"
            }
        ],
        "lamellar_hg": [
            "LamellarFFHGModel",
            {
                "sld": "sld_tail",
                "sld_solvent": "sld_solvent",
                "sld_head": "sld_head",
                "length_tail": "t_length",
                "length_head": "h_thickness"
            }
        ],
        "lamellar_hg_stack_caille": [
            "LamellarPSHGModel",
            {
                "sld": "sld_tail",
                "sld_head": "sld_head",
                "sld_solvent": "sld_solvent",
                "length_tail": "deltaT",
                "length_head": "deltaH",
                "d_spacing": "spacing",
                "Caille_parameter": "caille",
                "Nlayers": "n_plates",
            }
        ],
        "lamellar_stack_caille": [
            "LamellarPSModel",
            {
                "sld": "sld_bi",
                "sld_solvent": "sld_sol",
                "thickness": "delta",
                "d_spacing": "spacing",
                "Caille_parameter": "caille",
                "Nlayers": "n_plates",
            }
        ],
        "lamellar_stack_paracrystal": [
            "LamellarPCrystalModel",
            {
                "sld": "sld_layer",
                "sld_solvent": "sld_solvent",
                "thickness": "thickness",
                "d_spacing": "spacing",
                "sigma_d": "pd_spacing",
                "Nlayers": "Nlayers",
            }
        ],
        "line": [
            "LineModel",
            {
                "slope": "B",
                "scale": None,
                "background": None,
                "intercept": "A"
            }
        ],
        "linear_pearls": [
            "LinearPearlsModel",
            {
                "sld": "sld_pearl",
                "sld_solvent": "sld_solv",
                "edge_sep": "edge_separation"
            }
        ],
        "lorentz": [
            "Lorentz",
            {
                "cor_length": "length"
            },
            "LorentzModel",
        ],
        "mass_fractal": [
            "MassFractalModel",
            {
                "cutoff_length": "co_length",
                "radius": "radius",
                "fractal_dim_mass": "mass_dim"
            }
        ],
        "mass_surface_fractal": [
            "MassSurfaceFractal",
            {
                "rg_cluster": "cluster_rg",
                "fractal_dim_mass": "mass_dim",
                "radius": "radius",
                "fractal_dim_surf": "surface_dim",
                "rg_primary": "primary_rg"
            }
        ],
        "mono_gauss_coil": [
            "Debye",
            {
                "rg": "rg",
                "i_zero": "scale",
                "background": "background",
            },
            "DebyeModel",
        ],
        "multilayer_vesicle": [
            "MultiShellModel",
            {
                "radius": "core_radius",
                "sld_solvent": "core_sld",
                "n_shells": "n_pairs",
                "thick_shell": "s_thickness",
                "sld": "shell_sld",
                "thick_solvent": "w_thickness",
            }
        ],
        "onion": [
            "OnionExpShellModel",
            {
                "n_shells": "n_shells",
                "A": "A_shell",
                "sld_core": "sld_core0",
                "radius_core": "rad_core0",
                "sld_solvent": "sld_solv",
                "thickness": "thick_shell",
                "sld_in": "sld_in_shell",
                "sld_out": "sld_out_shell"
            }
        ],
        "parallelepiped": [
            "ParallelepipedModel",
            {
                "phi": "parallel_phi",
                "psi": "parallel_psi",
                "sld_solvent": "sldSolv",
                "length_a": "short_a",
                "length_b": "short_b",
                "sld": "sldPipe",
                "theta": "parallel_theta",
                "length_c": "long_c",
                "M0:sld": "M0_sld_pipe",
                "mtheta:sld": "M_theta_pipe",
                "mphi:sld": "M_phi_pipe",
                "M0:sld_solvent": "M0_sld_solv",
                "mtheta:sld_solvent": "M_theta_solv",
                "mphi:sld_solvent": "M_phi_solv",
                "up:frac_i": "up_frac_i",
                "up:frac_f": "up_frac_f",
                "up:angle": "up_angle",
            }
        ],
        "peak_lorentz": [
            "Peak Lorentz Model",
            {
                "peak_pos": "q0",
                "peak_hwhm": "B"
            },
            "PeakLorentzModel",
        ],
        "pearl_necklace": [
            "PearlNecklaceModel",
            {
                "scale": "scale",
                "thick_string": "thick_string",
                "sld_string": "sld_string",
                "sld_solvent": "sld_solv",
                "edge_sep": "edge_separation",
                "num_pearls": "num_pearls",
                "radius": "radius",
                "background": "background",
                "sld": "sld_pearl"
            }
        ],
        "poly_gauss_coil": [
            "Poly_GaussCoil",
            {
                "rg": "rg",
                "polydispersity": "poly_m",
                "i_zero": "scale",
                "background": "background",
            }
        ],
        "polymer_excl_volume": [
            "PolymerExclVolume",
            {
                "rg": "rg",
                "scale": "scale",
                "background": "background",
                "porod_exp": "m"
            }
        ],
        "polymer_micelle": [
            "MicelleSphCoreModel",
            {
                "sld_corona": "rho_corona",
                "sld_solvent": "rho_solv",
                "sld_core": "rho_core",
                "ndensity": "ndensity",
                "v_core": "v_core",
                "v_corona": "v_corona",
                "radius_core": "radius_core",
                "rg": "radius_gyr",
                "d_penetration": "d_penetration",
                "n_aggreg": "n_aggreg",
            }
        ],
        "porod": [
            "PorodModel",
            {
                "scale": "scale",
                "background": "background"
            }
        ],
        "power_law": [
            "PowerLawAbsModel",
            {
                "scale": "scale",
                "background": "background",
                "power": "m"
            }
        ],
        "pringle": [
            "PringlesModel",
            {
                "scale": "scale",
                "sld_solvent": "sld_solvent",
                "thickness": "thickness",
                "beta": "beta",
                "radius": "radius",
                "background": "background",
                "alpha": "alpha",
                "sld": "sld_pringle"
            }
        ],
        "raspberry": [
            "RaspBerryModel",
            {
                "volfraction_lg": "volf_Lsph",
                "volfraction_sm": "volf_Ssph",
                "radius_sm": "radius_Ssph",
                "radius_lg": "radius_Lsph",
                "sld_lg": "sld_Lsph",
                "sld_sm": "sld_Ssph",
                "sld_solvent": "sld_solv",
                "surface_fraction": "surfrac_Ssph",
                "penetration": "delta_Ssph"
            }
        ],
        "rectangular_prism": [
            "RectangularPrismModel",
            {
                "sld": "sldPipe",
                "length_a": "short_side",
                "b2a_ratio": "b2a_ratio",
                "c2a_ratio": "c2a_ratio",
                "sld_solvent": "sldSolv"
            }
        ],
        "rpa": [
            "RPA10Model",
            {
                "K12": "Kab", "K13": "Kac", "K14": "Kad",
                "K23": "Kbc", "K24": "Kbd", "K34": "Kcd",
                "N1": "Na", "N2": "Nb", "N3": "Nc", "N4": "Nd",
                "L1": "La", "L2": "Lb", "L3": "Lc", "L4": "Ld",
                "v1": "va", "v2": "vb", "v3": "vc", "v4": "vd",
                "b1": "ba", "b2": "bb", "b3": "bc", "b4": "bd",
                "Phi1": "Phia", "Phi2": "Phib", "Phi3": "Phic", "Phi4": "Phid",
                "case_num": "lcase_n"
            }
        ],
        "sc_paracrystal": [
            "SCCrystalModel",
            {
                "sld": "sldSph",
                "sld_solvent": "sldSolv"
            }
        ],
        "sphere": [
            "SphereModel",
            {
                "sld": "sldSph",
                "radius": "radius",
                "sld_solvent": "sldSolv",
                "M0:sld": "M0_sld_sph",
                "mtheta:sld": "M_theta_sph",
                "mphi:sld": "M_phi_sph",
                "M0:sld_solvent": "M0_sld_solv",
                "mtheta:sld_solvent": "M_theta_solv",
                "mphi:sld_solvent": "M_phi_solv",
                "up:frac_i": "up_frac_i",
                "up:frac_f": "up_frac_f",
                "up:angle": "up_angle"
            }
        ],
        "spherical_sld": [
            "SphericalSLDModel",
            # Be lazy and use a generator expression to define
            #    sld1: sld_flat0, ...
            #    thickness1: thick_flat0, ...
            #    interface1: thick_inter0, ...
            #    shape1: func_inter0, ...
            #    nu1: nu_inter0, ...
            # but override thickness1 => rad_cor0 and sld1 => sld_core0.
            # Note: explicit key,value pairs given by **{...} override the
            # keys from the gnerator expression ((k,v) for k,v in seq) when
            # used as dict((generator), **{...})
            dict(((field_new+str(index+1), field_old+str(index))
                  for field_new, field_old in [("sld", "sld_flat"),
                                               ("thickness", "thick_flat"),
                                               ("interface", "thick_inter"),
                                               ("shape", "func_inter"),
                                               ("nu", "nu_inter"),]
                  for index in range(11)),
                 **{
                     "n_shells": "n_shells",
                     "n_steps": "npts_inter",
                     "sld_solvent": "sld_solv",
                     "thickness1": "rad_core0",
                     "sld1": "sld_core0",
                 })
        ],
        "squarewell": [
            "SquareWellStructure",
            {
                "scale": "scale_factor",
                "radius_effective": "effect_radius",
                "wellwidth": "wellwidth",
                "welldepth": "welldepth",
            }
        ],
        "stacked_disks": [
            "StackedDisksModel",
            {
                "phi": "axis_phi",
                "sld_layer": "layer_sld",
                "sld_core": "core_sld",
                "theta": "axis_theta",
                "sld_solvent": "solvent_sld",
                "n_stacking": "n_stacking",
                "thick_layer": "layer_thick",
                "thick_core": "core_thick",
            }
        ],
        "star_polymer": [
            "StarPolymer",
            {
                "arms": "arms",
                "rg_squared": "R2"
            }
        ],
        "stickyhardsphere": [
            "StickyHSStructure",
            {
                "scale": "scale_factor",
                "radius_effective": "effect_radius",
            }
        ],
        "surface_fractal": [
            "SurfaceFractalModel",
            {
                "cutoff_length": "co_length",
                "radius": "radius",
                "fractal_dim_surf": "surface_dim"
            }
        ],
        "teubner_strey": [
            "TeubnerStreyModel",
            {
                # Note: parameters are completely rewritten in convert.py
                "volfraction_a": "volfraction_a",
                "sld_a": "sld_a",
                "sld_b": "sld_b",
                "d": "d",
                "xi": "xi",
            }
        ],
        "triaxial_ellipsoid": [
            "TriaxialEllipsoidModel",
            {
                "phi": "axis_phi",
                "radius_equat_minor": "semi_axisA",
                "radius_polar": "semi_axisC",
                "radius_equat_major": "semi_axisB",
                "sld_solvent": "sldSolv",
                "psi": "axis_psi",
                "sld": "sldEll",
                "theta": "axis_theta"
            }
        ],
        "two_lorentzian": [
            "TwoLorentzian",
            {
                "lorentz_scale_1": "scale_1",
                "lorentz_scale_2": "scale_2",
                "lorentz_exp_1": "exponent_1",
                "lorentz_exp_2": "exponent_2",
                "lorentz_length_2": "length_2",
                "lorentz_length_1": "length_1",
                "background": "background"
            },
            "TwoLorentzianModel",
        ],
        "two_power_law": [
            "TwoPowerLaw",
            {
                "coefficent_1": "coef_A",
                "power_2": "power2",
                "power_1": "power1",
                "background": "background",
                "crossover": "qc"
            },
            "TwoPowerLawModel",
        ],
        "unified_power_Rg": [
            "UnifiedPowerRg",
            dict(((field_new+str(index), field_old+str(index))
                  for field_new, field_old in [("rg", "Rg"),
                                               ("power", "power"),
                                               ("G", "G"),
                                               ("B", "B"),]
                  for index in range(11)),
                 **{
                     "background": "background",
                     "scale": "scale",
                 }),
            "UnifiedPowerRgModel",
        ],
        "vesicle": [
            "VesicleModel",
            {
                "sld": "shell_sld",
                "sld_solvent": "solv_sld"
            }
        ]
    },
    (5, 0, 4) : {
        "broad_peak": [
            "broad_peak",
            {
                "peak_scale": "lorentz_scale",
                "correlation_length": "lorentz_length",
                "shape_exp": None,
                "width_exp": "lorentz_exp",
            }
        ]
    }
}
